# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import

import numpy as np
from ..tools.chemistry import process_solution_chemistry
from .. import constants


def _ideal_activities_fct(molar_fractions, endmember_occupancies, n_endmembers,
                          n_occupancies, site_multiplicities,
                          endmember_configurational_entropies):
    site_occupancies = np.dot(molar_fractions, endmember_occupancies)
    a = np.power(site_occupancies,
                 endmember_occupancies * site_multiplicities).prod(-1)
    normalisation_constants = np.exp(endmember_configurational_entropies
                                     / constants.gas_constant)
    return normalisation_constants * a


def _non_ideal_hessian_fct(phi, molar_fractions, n_endmembers, alpha, W):
    q = np.eye(n_endmembers) - phi*np.ones((n_endmembers, n_endmembers))
    sum_pa = np.dot(molar_fractions, alpha)
    hess = np.einsum('m, i, ij, jk, mk->im', -alpha/sum_pa, -alpha, q, W, q)
    hess += hess.T
    return hess


def _non_ideal_interactions_fct(phi, molar_fractions, n_endmembers, alpha, W):
    # -sum(sum(qi.qj.Wij*)
    # equation (2) of Holland and Powell 2003
    q = np.eye(n_endmembers) - phi*np.ones((n_endmembers, n_endmembers))
    # The following are equivalent to
    # np.einsum('i, ij, jk, ik->i', -self.alphas, q, self.Wx, q)
    Wint = -alpha * (q.dot(W)*q).sum(-1)
    return Wint


def _non_ideal_hessian_subreg(p, n_endmembers, Wijk):
    Id = np.identity(n_endmembers)
    IIp = np.einsum('il, jm, k->ijklm', Id, Id, p)
    Ipp = np.einsum('il, j, k->ijkl', Id, p, p)
    ppp = np.einsum('i, j, k->ijk', p, p, p)

    A = (IIp
         + np.transpose(IIp, axes=[0, 2, 1, 3, 4])
         + np.transpose(IIp, axes=[1, 0, 2, 3, 4])
         + np.transpose(IIp, axes=[1, 2, 0, 3, 4])
         + np.transpose(IIp, axes=[2, 1, 0, 3, 4])
         + np.transpose(IIp, axes=[2, 0, 1, 3, 4]))
    B = 2.*(Ipp
            + np.transpose(Ipp, axes=[1, 0, 2, 3])
            + np.transpose(Ipp, axes=[2, 1, 0, 3]))

    Asum = (A
            - B[:, :, :, :, None]
            - B[:, :, :, None, :]
            + 6.*ppp[:, :, :, None, None])
    hess = np.einsum('ijklm, ijk->lm', Asum, Wijk)
    return hess


def _non_ideal_interactions_subreg(p, n_endmembers, Wijk):
    Aijkl = np.einsum('li, j, k->ijkl', np.identity(n_endmembers), p, p)
    ppp = np.einsum('i, j, k->ijk', p, p, p)

    Asum = (Aijkl
            + np.transpose(Aijkl, axes=[1, 0, 2, 3])
            + np.transpose(Aijkl, axes=[1, 2, 0, 3])
            - 2*ppp[:, :, :, None])

    Wint = np.einsum('ijk, ijkl->l', Wijk, Asum)
    return Wint


def logish(x, eps=1.e-5):
    """
    2nd order series expansion of log(x) about eps:
    log(eps) - sum_k=1^infty (f_eps)^k / k
    Prevents infinities at x=0
    """
    f_eps = 1. - x/eps
    mask = x > eps
    ln = np.where(x <= eps, np.log(eps) - f_eps - f_eps*f_eps/2., 0.)
    ln[mask] = np.log(x[mask])
    return ln


def inverseish(x, eps=1.e-5):
    """
    1st order series expansion of 1/x about eps: 2/eps - x/eps/eps
    Prevents infinities at x=0
    """
    mask = x > eps
    oneoverx = np.where(x <= eps, 2./eps - x/eps/eps, 0.)
    oneoverx[mask] = 1./x[mask]
    return oneoverx


class SolutionModel(object):

    """
    This is the base class for a solution model,  intended for use
    in defining solid solutions and performing thermodynamic calculations
    on them.  All minerals of type :class:`burnman.SolidSolution` use
    a solution model for defining how the endmembers in the solid solution
    interact.

    A user wanting a new solution model should define the functions included
    in the base class. All of the functions in the base class return zero,
    so if the user-defined solution model does not implement them,
    they essentially have no effect, and the Gibbs free energy and molar
    volume of a solid solution will be equal to the weighted arithmetic
    averages of the different endmember values.
    """

    def __init__(self):
        """
        Does nothing.
        """
        pass

    def excess_gibbs_free_energy(self, pressure, temperature, molar_fractions):
        """
        Given a list of molar fractions of different phases,
        compute the excess Gibbs free energy of the solution.
        The base class implementation assumes that the excess gibbs
        free energy is zero.

        Parameters
        ----------
        pressure : float
            Pressure at which to evaluate the solution model. [Pa]

        temperature : float
            Temperature at which to evaluate the solution. [K]

        molar_fractions : list of floats
            List of molar fractions of the different endmembers in solution

        Returns
        -------
        G_excess : float
            The excess Gibbs free energy
        """
        return np.dot(np.array(molar_fractions), self.excess_partial_gibbs_free_energies(pressure, temperature, molar_fractions))

    def excess_volume(self, pressure, temperature, molar_fractions):
        """
        Given a list of molar fractions of different phases,
        compute the excess volume of the solution.
        The base class implementation assumes that the excess volume is zero.

        Parameters
        ----------
        pressure : float
            Pressure at which to evaluate the solution model. [Pa]

        temperature : float
            Temperature at which to evaluate the solution. [K]

        molar_fractions : list of floats
            List of molar fractions of the different endmembers in solution

        Returns
        -------
        V_excess : float
            The excess volume of the solution
        """
        return np.dot(molar_fractions,
                      self.excess_partial_volumes(pressure, temperature, molar_fractions))

    def excess_entropy(self, pressure, temperature, molar_fractions):
        """
        Given a list of molar fractions of different phases,
        compute the excess entropy of the solution.
        The base class implementation assumes that the excess entropy is zero.

        Parameters
        ----------
        pressure : float
            Pressure at which to evaluate the solution model. [Pa]

        temperature : float
            Temperature at which to evaluate the solution. [K]

        molar_fractions : list of floats
            List of molar fractions of the different endmembers in solution

        Returns
        -------
        S_excess : float
            The excess entropy of the solution
        """
        return np.dot(molar_fractions,
                      self.excess_partial_entropies(pressure, temperature, molar_fractions))

    def excess_enthalpy(self, pressure, temperature, molar_fractions):
        """
        Given a list of molar fractions of different phases,
        compute the excess enthalpy of the solution.
        The base class implementation assumes that the excess enthalpy is zero.

        Parameters
        ----------
        pressure : float
            Pressure at which to evaluate the solution model. [Pa]

        temperature : float
            Temperature at which to evaluate the solution. [K]

        molar_fractions : list of floats
            List of molar fractions of the different endmembers in solution

        Returns
        -------
        H_excess : float
            The excess enthalpy of the solution
        """
        return (self.excess_gibbs_free_energy(pressure, temperature, molar_fractions)
                + temperature*self.excess_entropy(pressure, temperature, molar_fractions))

    def excess_partial_gibbs_free_energies(self, pressure, temperature, molar_fractions):
        """
        Given a list of molar fractions of different phases,
        compute the excess Gibbs free energy for each endmember of the solution.
        The base class implementation assumes that the excess gibbs
        free energy is zero.

        Parameters
        ----------
        pressure : float
            Pressure at which to evaluate the solution model. [Pa]

        temperature : float
            Temperature at which to evaluate the solution. [K]

        molar_fractions : list of floats
            List of molar fractions of the different endmembers in solution

        Returns
        -------
        partial_G_excess : numpy array
            The excess Gibbs free energy of each endmember
        """
        return np.zeros_like(molar_fractions)

    def excess_partial_entropies(self, pressure, temperature, molar_fractions):
        """
        Given a list of molar fractions of different phases,
        compute the excess entropy for each endmember of the solution.
        The base class implementation assumes that the excess entropy
        is zero (true for mechanical solutions).

        Parameters
        ----------
        pressure : float
            Pressure at which to evaluate the solution model. [Pa]

        temperature : float
            Temperature at which to evaluate the solution. [K]

        molar_fractions : list of floats
            List of molar fractions of the different endmembers in solution

        Returns
        -------
        partial_S_excess : numpy array
            The excess entropy of each endmember
        """
        return np.zeros_like(molar_fractions)

    def excess_partial_volumes(self, pressure, temperature, molar_fractions):
        """
        Given a list of molar fractions of different phases,
        compute the excess volume for each endmember of the solution.
        The base class implementation assumes that the excess volume
        is zero.

        Parameters
        ----------
        pressure : float
            Pressure at which to evaluate the solution model. [Pa]

        temperature : float
            Temperature at which to evaluate the solution. [K]

        molar_fractions : list of floats
            List of molar fractions of the different endmembers in solution

        Returns
        -------
        partial_V_excess : numpy array
            The excess volume of each endmember
        """
        return np.zeros_like(np.array(molar_fractions))


class MechanicalSolution (SolutionModel):

    """
    An extremely simple class representing a mechanical solution model.
    A mechanical solution experiences no interaction between endmembers.
    Therefore, unlike ideal solutions there is no entropy of mixing;
    the total gibbs free energy of the solution is equal to the
    dot product of the molar gibbs free energies and molar fractions
    of the constituent materials.
    """

    def __init__(self, endmembers):
        self.n_endmembers = len(endmembers)
        self.formulas = [e[1] for e in endmembers]

    def excess_gibbs_free_energy(self, pressure, temperature, molar_fractions):
        return 0.

    def excess_volume(self, pressure, temperature, molar_fractions):
        return 0.

    def excess_entropy(self, pressure, temperature, molar_fractions):
        return 0.

    def excess_enthalpy(self, pressure, temperature, molar_fractions):
        return 0.

    def excess_partial_gibbs_free_energies(self, pressure, temperature, molar_fractions):
        return np.zeros_like(molar_fractions)

    def excess_partial_volumes(self, pressure, temperature, molar_fractions):
        return np.zeros_like(molar_fractions)

    def excess_partial_entropies(self, pressure, temperature, molar_fractions):
        return np.zeros_like(molar_fractions)

    def activity_coefficients(self, pressure, temperature, molar_fractions):
        return np.ones_like(molar_fractions)

    def activities(self, pressure, temperature, molar_fractions):
        return np.ones_like(molar_fractions)


class IdealSolution (SolutionModel):

    """
    A very simple class representing an ideal solution model.
    Calculate the excess gibbs free energy and entropy due to configurational
    entropy, excess volume is equal to zero.
    """

    def __init__(self, endmembers):
        self.n_endmembers = len(endmembers)
        self.formulas = [e[1] for e in endmembers]

        # Process solid solution chemistry
        process_solution_chemistry(self)

        self._calculate_endmember_configurational_entropies()

    def _calculate_endmember_configurational_entropies(self):
        S_conf = -constants.gas_constant * (self.site_multiplicities
                                            * self.endmember_occupancies
                                            * logish(self.endmember_occupancies)).sum(-1)
        self.endmember_configurational_entropies = S_conf

    def excess_partial_gibbs_free_energies(self, pressure, temperature, molar_fractions):
        return self._ideal_excess_partial_gibbs(temperature, molar_fractions)

    def excess_partial_entropies(self, pressure, temperature, molar_fractions):
        return self._ideal_excess_partial_entropies(temperature, molar_fractions)

    def excess_partial_volumes(self, pressure, temperature, molar_fractions):
        return np.zeros((self.n_endmembers))

    def gibbs_hessian(self, pressure, temperature, molar_fractions):
        hess_S = self._ideal_entropy_hessian(temperature, molar_fractions)
        return -temperature*hess_S

    def entropy_hessian(self, pressure, temperature, molar_fractions):
        hess_S = self._ideal_entropy_hessian(temperature, molar_fractions)
        return hess_S

    def volume_hessian(self, pressure, temperature, molar_fractions):
        return np.zeros((len(molar_fractions), len(molar_fractions)))

    def _configurational_entropy(self, molar_fractions):
        site_occupancies = np.dot(molar_fractions, self.endmember_occupancies)
        conf_entropy = - constants.gas_constant * (site_occupancies
                                                   * self.site_multiplicities
                                                   * logish(site_occupancies)).sum(-1)
        return conf_entropy

    def _ideal_excess_partial_gibbs(self, temperature, molar_fractions):
        return -temperature*self._ideal_excess_partial_entropies(temperature, molar_fractions)

    def _ideal_excess_partial_entropies(self, temperature, molar_fractions):
        return -constants.gas_constant * self._log_ideal_activities(molar_fractions)

    def _ideal_entropy_hessian(self, temperature, molar_fractions):
        hessian = -constants.gas_constant * self._log_ideal_activity_derivatives(molar_fractions)
        return hessian

    def _log_ideal_activities(self, molar_fractions):
        site_occupancies = np.dot(molar_fractions, self.endmember_occupancies)
        lna = (self.endmember_occupancies * self.site_multiplicities
               * logish(site_occupancies)).sum(-1)
        normalisation_constants = (self.endmember_configurational_entropies
                                   / constants.gas_constant)

        return lna + normalisation_constants

    def _log_ideal_activity_derivatives(self, molar_fractions):
        site_occupancies = np.dot(molar_fractions, self.endmember_occupancies)
        n_site_atoms = self.endmember_occupancies[0].dot(self.site_multiplicities)
        dlnadp = np.einsum('ij, j, mj',
                           self.endmember_occupancies,
                           self.site_multiplicities*inverseish(site_occupancies),
                           self.endmember_occupancies) - n_site_atoms
        return dlnadp

    def _ideal_activities(self, molar_fractions):
        return _ideal_activities_fct(molar_fractions,
                                     self.endmember_occupancies,
                                     self.n_endmembers,
                                     self.n_occupancies,
                                     self.site_multiplicities,
                                     self.endmember_configurational_entropies)

    def activity_coefficients(self, pressure, temperature, molar_fractions):
        return np.ones_like(molar_fractions)

    def activities(self, pressure, temperature, molar_fractions):
        return self._ideal_activities(molar_fractions)


class AsymmetricRegularSolution (IdealSolution):

    """
    Solution model implementing the asymmetric regular solution model
    formulation as described in :cite:`HP2003`.

    The excess nonconfigurational Gibbs energy is given by the
    expression:

    .. math::
        \\mathcal{G}_{\\textrm{excess}} = \\alpha^T p (\\phi^T W \\phi)

    :math:`\\alpha` is a vector of van Laar parameters governing asymmetry
    in the excess properties.

    .. math::
        \\phi_i = \\frac{\\alpha_i p_i}{\\sum_{k=1}^{n} \\alpha_k p_k},
        W_{ij} = \\frac{2 w_{ij}}{\\alpha_i + \\alpha_j} \\textrm{for i<j}
    """

    def __init__(self, endmembers, alphas, energy_interaction,
                 volume_interaction=None, entropy_interaction=None):

        self.n_endmembers = len(endmembers)

        # Create array of van Laar parameters
        self.alphas = np.array(alphas)

        # Create 2D arrays of interaction parameters
        self.We = np.triu(2. / (self.alphas[:, np.newaxis] + self.alphas), 1)
        self.We[np.triu_indices(self.n_endmembers, 1)] *= np.array([i for row in energy_interaction
                                                                    for i in row])

        if entropy_interaction is not None:
            self.Ws = np.triu(2. / (self.alphas[:, np.newaxis] + self.alphas), 1)
            self.Ws[np.triu_indices(self.n_endmembers, 1)] *= np.array([i for row in entropy_interaction
                                                                        for i in row])
        else:
            self.Ws = np.zeros((self.n_endmembers, self.n_endmembers))

        if volume_interaction is not None:
            self.Wv = np.triu(2. / (self.alphas[:, np.newaxis] + self.alphas), 1)
            self.Wv[np.triu_indices(self.n_endmembers, 1)] *= np.array([i for row in volume_interaction
                                                                        for i in row])
        else:
            self.Wv = np.zeros((self.n_endmembers, self.n_endmembers))

        # initialize ideal solution model
        IdealSolution.__init__(self, endmembers)

    def _phi(self, molar_fractions):
        phi = self.alphas * molar_fractions
        phi = np.divide(phi, np.sum(phi))
        return phi

    def _non_ideal_interactions(self, W, molar_fractions):
        # -sum(sum(qi.qj.Wij*)
        # equation (2) of Holland and Powell 2003
        phi = self._phi(molar_fractions)
        return _non_ideal_interactions_fct(phi, np.array(molar_fractions), self.n_endmembers, self.alphas, W)

    def _non_ideal_excess_partial_gibbs(self, pressure, temperature, molar_fractions):
        Eint = self._non_ideal_interactions(self.We, molar_fractions)
        Sint = self._non_ideal_interactions(self.Ws, molar_fractions)
        Vint = self._non_ideal_interactions(self.Wv, molar_fractions)
        return Eint - temperature * Sint + pressure * Vint

    def excess_partial_gibbs_free_energies(self, pressure, temperature, molar_fractions):
        ideal_gibbs = IdealSolution._ideal_excess_partial_gibbs(
            self, temperature, molar_fractions)
        non_ideal_gibbs = self._non_ideal_excess_partial_gibbs(
            pressure, temperature, molar_fractions)
        return ideal_gibbs + non_ideal_gibbs

    def excess_partial_entropies(self, pressure, temperature, molar_fractions):
        ideal_entropies = IdealSolution._ideal_excess_partial_entropies(
            self, temperature, molar_fractions)
        non_ideal_entropies = self._non_ideal_interactions(self.Ws, molar_fractions)
        return ideal_entropies + non_ideal_entropies

    def excess_partial_volumes(self, pressure, temperature, molar_fractions):
        return self._non_ideal_interactions(self.Wv, molar_fractions)

    def gibbs_hessian(self, pressure, temperature, molar_fractions):
        ideal_entropy_hessian = IdealSolution._ideal_entropy_hessian(self, temperature, molar_fractions)
        phi = self._phi(molar_fractions)
        nonideal_gibbs_hessian = _non_ideal_hessian_fct(phi, molar_fractions,
                                                        self.n_endmembers, self.alphas,
                                                        self.We - temperature*self.Ws + pressure*self.Wv)

        return nonideal_gibbs_hessian - temperature*ideal_entropy_hessian

    def entropy_hessian(self, pressure, temperature, molar_fractions):
        ideal_entropy_hessian = IdealSolution._ideal_entropy_hessian(self, temperature, molar_fractions)
        phi = self._phi(molar_fractions)
        nonideal_entropy_hessian = _non_ideal_hessian_fct(phi, molar_fractions,
                                                          self.n_endmembers, self.alphas,
                                                          self.Ws)
        return ideal_entropy_hessian + nonideal_entropy_hessian

    def volume_hessian(self, pressure, temperature, molar_fractions):
        phi = self._phi(molar_fractions)
        return _non_ideal_hessian_fct(phi, molar_fractions,
                                      self.n_endmembers, self.alphas,
                                      self.Wv)

    def activity_coefficients(self, pressure, temperature, molar_fractions):
        if temperature > 1.e-10:
            return np.exp(self._non_ideal_excess_partial_gibbs(pressure, temperature, molar_fractions) / (constants.gas_constant * temperature))
        else:
            raise Exception("Activity coefficients not defined at 0 K.")

    def activities(self, pressure, temperature, molar_fractions):
        return IdealSolution.activities(self, pressure, temperature, molar_fractions) * self.activity_coefficients(pressure, temperature, molar_fractions)


class SymmetricRegularSolution (AsymmetricRegularSolution):

    """
    Solution model implementing the symmetric regular solution model.
    This is a special case of the
    :class:`burnman.solutionmodel.AsymmetricRegularSolution` class.
    """

    def __init__(self, endmembers, energy_interaction, volume_interaction=None, entropy_interaction=None):
        alphas = np.ones(len(endmembers))
        AsymmetricRegularSolution.__init__(
            self, endmembers, alphas, energy_interaction, volume_interaction, entropy_interaction)


class SubregularSolution (IdealSolution):

    """
    Solution model implementing the subregular solution model formulation
    as described in :cite:`HW1989`. The excess conconfigurational
    Gibbs energy is given by the expression:

    .. math::
        \\mathcal{G}_{\\textrm{excess}} = \\sum_i \\sum_{j > i} (p_i p_j^2
        W_{ij} + p_j p_i^2 W_{ji} + \\sum_{k > j > i} p_i p_j p_k W_{ijk})

    Interaction parameters are inserted into a 3D interaction matrix during
    initialization to make use of numpy vector algebra.

    Parameters
    ----------
    endmembers : list of lists
        A list of all the independent endmembers in the solution.
        The first item of each list gives the Mineral object corresponding
        to the endmember. The second item gives the site-species formula.

    energy_interaction : list of list of lists
        The binary endmember interaction energies.
        Each interaction[i, j-i-1, 0] corresponds to W(i,j), while
        interaction[i, j-i-1, 1] corresponds to W(j,i).

    volume_interaction : list of list of lists
        The binary endmember interaction volumes.
        Each interaction[i, j-i-1, 0] corresponds to W(i,j), while
        interaction[i, j-i-1, 1] corresponds to W(j,i).

    entropy_interaction : list of list of lists
        The binary endmember interaction entropies.
        Each interaction[i, j-i-1, 0] corresponds to W(i,j), while
        interaction[i, j-i-1, 1] corresponds to W(j,i).

    energy_ternary_terms : list of lists
        The ternary interaction energies. Each list should contain
        four entries: the indices i, j, k and the value of the interaction.

    volume_ternary_terms : list of lists
        The ternary interaction volumes. Each list should contain
        four entries: the indices i, j, k and the value of the interaction.

    entropy_ternary_terms : list of lists
        The ternary interaction entropies. Each list should contain
        four entries: the indices i, j, k and the value of the interaction.
    """

    def __init__(self, endmembers, energy_interaction,
                 volume_interaction=None, entropy_interaction=None,
                 energy_ternary_terms=None,
                 volume_ternary_terms=None, entropy_ternary_terms=None):
        """
        Initialization function for the SubregularSolution class.
        """

        self.n_endmembers = len(endmembers)

        # Create 3D arrays of interaction parameters
        self.Wijke = np.zeros(shape=(self.n_endmembers,
                                     self.n_endmembers,
                                     self.n_endmembers))
        self.Wijks = np.zeros_like(self.Wijke)
        self.Wijkv = np.zeros_like(self.Wijke)

        # setup excess enthalpy interaction matrix
        for i in range(self.n_endmembers):
            for j in range(i + 1, self.n_endmembers):
                w0 = energy_interaction[i][j - i - 1][0]/2.
                w1 = energy_interaction[i][j - i - 1][1]/2.
                self.Wijke[:, i, j] += w0
                self.Wijke[:, j, i] += w1

                self.Wijke[i, j, j] += w0
                self.Wijke[j, i, i] += w1

                self.Wijke[i, j, i] -= w0
                self.Wijke[j, i, j] -= w1

        if energy_ternary_terms is not None:
            for (i, j, k, v) in energy_ternary_terms:
                self.Wijke[i, j, k] += v

        if entropy_interaction is not None:
            for i in range(self.n_endmembers):
                for j in range(i + 1, self.n_endmembers):
                    w0 = entropy_interaction[i][j - i - 1][0]/2.
                    w1 = entropy_interaction[i][j - i - 1][1]/2.
                    self.Wijks[:, i, j] += w0
                    self.Wijks[:, j, i] += w1

                    self.Wijks[i, j, j] += w0
                    self.Wijks[j, i, i] += w1

                    self.Wijks[i, j, i] -= w0
                    self.Wijks[j, i, j] -= w1

        if entropy_ternary_terms is not None:
            for (i, j, k, v) in entropy_ternary_terms:
                self.Wijks[i, j, k] += v

        if volume_interaction is not None:
            for i in range(self.n_endmembers):
                for j in range(i + 1, self.n_endmembers):
                    w0 = volume_interaction[i][j - i - 1][0]/2.
                    w1 = volume_interaction[i][j - i - 1][1]/2.
                    self.Wijkv[:, i, j] += w0
                    self.Wijkv[:, j, i] += w1

                    self.Wijkv[i, j, j] += w0
                    self.Wijkv[j, i, i] += w1

                    self.Wijkv[i, j, i] -= w0
                    self.Wijkv[j, i, j] -= w1

        if volume_ternary_terms is not None:
            for (i, j, k, v) in volume_ternary_terms:
                self.Wijkv[i, j, k] += v

        # initialize ideal solution model
        IdealSolution.__init__(self, endmembers)

    def _non_ideal_function(self, Wijk, molar_fractions):
        n = len(molar_fractions)
        return _non_ideal_interactions_subreg(molar_fractions, n, Wijk)

    def _non_ideal_interactions(self, molar_fractions):
        # equation (6') of Helffrich and Wood, 1989
        Eint = self._non_ideal_function(self.Wijke, molar_fractions)
        Sint = self._non_ideal_function(self.Wijks, molar_fractions)
        Vint = self._non_ideal_function(self.Wijkv, molar_fractions)
        return Eint, Sint, Vint

    def _non_ideal_excess_partial_gibbs(self, pressure, temperature, molar_fractions):
        Eint, Sint, Vint = self._non_ideal_interactions(molar_fractions)
        return Eint - temperature * Sint + pressure * Vint

    def excess_partial_gibbs_free_energies(self, pressure, temperature, molar_fractions):
        ideal_gibbs = IdealSolution._ideal_excess_partial_gibbs(
            self, temperature, molar_fractions)
        non_ideal_gibbs = self._non_ideal_excess_partial_gibbs(
            pressure, temperature, molar_fractions)
        return ideal_gibbs + non_ideal_gibbs

    def excess_partial_entropies(self, pressure, temperature, molar_fractions):
        ideal_entropies = IdealSolution._ideal_excess_partial_entropies(
            self, temperature, molar_fractions)
        non_ideal_entropies = self._non_ideal_function(self.Wijks, molar_fractions)
        return ideal_entropies + non_ideal_entropies

    def excess_partial_volumes(self, pressure, temperature, molar_fractions):
        non_ideal_volumes = self._non_ideal_function(self.Wijkv, molar_fractions)
        return non_ideal_volumes

    def gibbs_hessian(self, pressure, temperature, molar_fractions):
        n = len(molar_fractions)
        ideal_entropy_hessian = IdealSolution._ideal_entropy_hessian(self, temperature, molar_fractions)
        nonideal_gibbs_hessian = _non_ideal_hessian_subreg(molar_fractions, n,
                                                           self.Wijke - temperature*self.Wijks
                                                           + pressure*self.Wijkv)

        return nonideal_gibbs_hessian - temperature*ideal_entropy_hessian

    def entropy_hessian(self, pressure, temperature, molar_fractions):
        n = len(molar_fractions)
        ideal_entropy_hessian = IdealSolution._ideal_entropy_hessian(self, temperature, molar_fractions)
        nonideal_entropy_hessian = _non_ideal_hessian_subreg(molar_fractions, n,
                                                             self.Wijks)
        return ideal_entropy_hessian + nonideal_entropy_hessian

    def volume_hessian(self, pressure, temperature, molar_fractions):
        n = len(molar_fractions)
        return _non_ideal_hessian_subreg(molar_fractions, n, self.Wijkv)

    def activity_coefficients(self, pressure, temperature, molar_fractions):
        if temperature > 1.e-10:
            return np.exp(self._non_ideal_excess_partial_gibbs(pressure, temperature, molar_fractions) / (constants.gas_constant * temperature))
        else:
            raise Exception("Activity coefficients not defined at 0 K.")

    def activities(self, pressure, temperature, molar_fractions):
        return IdealSolution.activities(self, pressure, temperature, molar_fractions) * self.activity_coefficients(pressure, temperature, molar_fractions)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
from Kurnosov et al.  Nature 2017
Kurnosov, A., et al. "Evidence for a Fe3+-rich pyrolitic lower mantle from (Al, Fe)-bearing bridgmanite elasticity data."
Nature 543.7646 (2017): 543-546. doi:10.1038/nature21390
"""

from __future__ import absolute_import

from ..classes.mineral import Mineral
from ..classes.solidsolution import SolidSolution
from ..classes.solutionmodel import *
from ..tools.chemistry import dictionarize_formula, formula_mass

"""
SOLID SOLUTIONS
"""


class bridgmanite(SolidSolution):
    def __init__(self, molar_fractions=None):
        self.name = 'bridgmanite/perovskite'
        self.solution_type = 'ideal'
        self.endmembers = [[mg_si_perovskite(), '[Mg][Si]O3'],
                           [fe_si_perovskite(), '[Fe][Si]O3'],
                           [al_al_perovskite(), '[Al][Al]O3'],
                           [fe_al_perovskite(), '[Fe][Al]O3']]
        SolidSolution.__init__(self, molar_fractions=molar_fractions)

class ferropericlase(SolidSolution):
    def __init__(self, molar_fractions=None):
        self.name = 'magnesiowustite/ferropericlase'
        self.solution_type = 'symmetric'
        self.endmembers = [[periclase(), '[Mg]O'],
                           [wuestite(), '[Fe]O']]
        self.energy_interaction = [[13.e3]]

        SolidSolution.__init__(self, molar_fractions=molar_fractions)

"""
ENDMEMBERS
"""

class mg_si_perovskite (Mineral):

    def __init__(self):
        formula = 'MgSiO3'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'MgSiO3 perovskite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -1368000.0,
            'V_0': 2.4445e-05,
            'K_0': 2.51e+11,
            'Kprime_0': 4.1,
            'Debye_0': 905.,
            'grueneisen_0': 1.57,
            'q_0': 1.1,
            'G_0': 1.73e+11,
            'Gprime_0': 1.7,
            'eta_s_0': 2.3,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        Mineral.__init__(self)

class fe_si_perovskite (Mineral):

    def __init__(self):
        formula = 'FeSiO3'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'FeSiO3 perovskite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -1043000.0,
            'V_0': 2.534e-05,
            'K_0': 2.72e+11,
            'Kprime_0': 4.1,
            'Debye_0': 871.,
            'grueneisen_0': 1.57,
            'q_0': 1.1,
            'G_0': 1.33+11,
            'Gprime_0': 1.4,
            'eta_s_0': 2.3,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        Mineral.__init__(self)


class fe_al_perovskite (Mineral):

    def __init__(self):
        formula = 'FeAlO3'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'FeAlO3 perovskite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'V_0': 2.69e-05,
            'K_0': 2.20e+11,
            'Kprime_0': 1.3,
            'Debye_0': 886.,
            'grueneisen_0': 1.57,
            'q_0': 1.1,
            'G_0': 96.0e+09,
            'Gprime_0': 3.4,
            'eta_s_0': 2.5,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        Mineral.__init__(self)


class al_al_perovskite(Mineral):

    def __init__(self):
        formula = 'Al2O3'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Al2O3 perovskite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -1533878.0,
            'V_0': 2.494e-05,
            'K_0': 2.58e+11,
            'Kprime_0': 4.1,
            'Debye_0': 886.,
            'grueneisen_0': 1.57,
            'q_0': 1.1,
            'G_0': 1.71e+11,
            'Gprime_0': 1.5,
            'eta_s_0': 2.5,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        Mineral.__init__(self)


class periclase (Mineral):

    def __init__(self):
        formula = 'MgO'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Periclase',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -569000.,
            'V_0': 1.124e-05,
            'K_0': 1.602e+11,
            'Kprime_0': 3.99,
            'Debye_0': 767.0,
            'grueneisen_0': 1.36,
            'q_0': 1.7,
            'G_0': 1.31e+11,
            'Gprime_0': 2.1,
            'eta_s_0': 2.8,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        Mineral.__init__(self)


class wuestite (Mineral):

    def __init__(self):
        formula = 'FeO'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Wuestite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -242000.0,
            'V_0': 1.226e-05,
            'K_0': 1.49e+11,
            'Kprime_0': 3.6,
            'Debye_0': 454.,
            'grueneisen_0': 1.53,
            'q_0': 1.7,
            'G_0': 60.0e9,
            'Gprime_0': 1.8,
            'eta_s_0': 0.6,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        Mineral.__init__(self)



class ca_perovskite (Mineral):

    def __init__(self):
        formula = 'CaSiO3'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'CaSiO3 perovskite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -1463358.0,
            'V_0': 2.754e-05,
            'K_0': 2.36e+11,
            'Kprime_0': 3.9,
            'Debye_0': 802.,
            'grueneisen_0': 1.89,
            'q_0': 0.9,
            'G_0': 1.57e+11,
            'Gprime_0': 2.2,
            'eta_s_0': 1.3,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}


        Mineral.__init__(self)



class hcp_iron (Mineral):

    def __init__(self):
        formula = 'Fe'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Hexagonal close packed iron',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -1463358.0,
            'V_0': 0.673e-05,
            'K_0': 1.64e+11,
            'Kprime_0': 4.,
            'Debye_0': 422.,
            'grueneisen_0': 1.71,
            'q_0': 1.4,
            'G_0': .815e+11,
            'Gprime_0': 1.9,
            'eta_s_0': 7,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}


        Mineral.__init__(self)
'''
Mineral aliases
'''
perovskite=bridgmanite



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

from . import equation_of_state as eos
import warnings
import numpy as np


def volume(pressure, V_0, K_0, Kprime_0):
    return V_0 * np.power(1. + (pressure * Kprime_0 / K_0), -1./Kprime_0)


def pressure(volume, V_0, K_0, Kprime_0):
    return K_0 / Kprime_0 * (np.power(volume / V_0, -Kprime_0) - 1.)


def bulk_modulus(pressure, K_0, Kprime_0):
    return K_0 + pressure * Kprime_0


def energy(volume, E_0, V_0, K_0, Kprime_0):
    Vrel = volume / V_0
    return (E_0 + K_0*V_0 * (np.power(Vrel, 1. - Kprime_0)
                             / (Kprime_0*(Kprime_0 - 1))
                             + Vrel / Kprime_0 - 1./(Kprime_0 - 1.)))


def intVdP(pressure, V_0, K_0, Kprime_0):
    return (V_0 * K_0
            * ((np.power(1. + (pressure * Kprime_0 / K_0),
                         1.-(1./Kprime_0))) - 1.)
            / (Kprime_0 - 1.))


class Murnaghan(eos.EquationOfState):

    """
    Base class for the isothermal Murnaghan equation of state,
    as described in :cite:`Murnaghan1944`.
    """

    def volume(self, pressure, temperature, params):
        """
        Returns volume :math:`[m^3]` as a function of pressure :math:`[Pa]`.
        """
        return volume(pressure,
                      params['V_0'], params['K_0'], params['Kprime_0'])

    def pressure(self, temperature, volume, params):
        return pressure(volume,
                        params['V_0'], params['K_0'], params['Kprime_0'])

    def isothermal_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus :math:`K_T` :math:`[Pa]`
        as a function of pressure :math:`[Pa]`,
        temperature :math:`[K]` and volume :math:`[m^3]`.
        """
        return bulk_modulus(pressure, params['K_0'], params['Kprime_0'])

    def adiabatic_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus :math:`K_s` of the mineral. :math:`[Pa]`.
        """
        return bulk_modulus(pressure, params['K_0'], params['Kprime_0'])

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Returns shear modulus :math:`G` of the mineral. :math:`[Pa]`
        Currently not included in the Murnghan EOS, so omitted.
        """
        return 0.

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the molar entropy :math:`\mathcal{S}` of the mineral.
        :math:`[J/K/mol]`
        """
        return 0.

    def molar_internal_energy(self, pressure, temperature, volume, params):
        """
        Returns the internal energy :math:`\mathcal{E}` of the mineral.
        :math:`[J/mol]`
        """
        return energy(volume,
                      params['E_0'], params['V_0'],
                      params['K_0'], params['Kprime_0'])

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Gibbs free energy :math:`\mathcal{G}` of the mineral.
        :math:`[J/mol]`
        """
        # G = E + PV
        return (self.molar_internal_energy(pressure, temperature,
                                           volume, params)
                + volume*pressure)

    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects,
        return a very large number. :math:`[J/K/mol]`
        """
        return 1.e99

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects,
        return a very large number. :math:`[J/K/mol]`
        """
        return 1.e99

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects,
        return zero. :math:`[1/K]`
        """
        return 0.

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects,
        return zero. :math:`[unitless]`
        """
        return 0.

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """

        if 'E_0' not in params:
            params['E_0'] = 0.
        if 'P_0' not in params:
            params['P_0'] = 0.

        # G is not included in the Murnaghan EOS so we shall set them to NaN's
        if 'G_0' not in params:
            params['G_0'] = float('nan')
        if 'Gprime_0' not in params:
            params['Gprime_0'] = float('nan')

        # check that all the required keys are in the dictionary
        expected_keys = ['V_0', 'K_0', 'Kprime_0']
        for k in expected_keys:
            if k not in params:
                raise KeyError('params object missing parameter : ' + k)

        # now check that the values are reasonable.  I mostly just
        # made up these values from experience, and we are only
        # raising a warning.  Better way to do this? [IR]
        if params['V_0'] < 1.e-7 or params['V_0'] > 1.e-3:
            warnings.warn('Unusual value for V_0', stacklevel=2)
        if params['K_0'] < 1.e9 or params['K_0'] > 1.e13:
            warnings.warn('Unusual value for K_0', stacklevel=2)
        if params['Kprime_0'] < -5. or params['Kprime_0'] > 30.:
            warnings.warn('Unusual value for Kprime_0', stacklevel=2)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import
from __future__ import print_function

import numpy as np
from scipy.ndimage.filters import gaussian_filter
from scipy.interpolate import interp2d
from sympy import Matrix, Rational
import scipy.integrate as integrate
from collections import Counter
import itertools

from .reductions import row_reduce


def round_to_n(x, xerr, n):
    return round(x, -int(np.floor(np.log10(np.abs(xerr)))) + (n - 1))


def unit_normalize(a, order=2, axis=-1):
    """
    Calculates the L2 normalized array of numpy array a
    of a given order and along a given axis.
    """
    l2 = np.atleast_1d(np.apply_along_axis(np.linalg.norm, axis, a, order))

    l2[l2 == 0] = 1
    return a / np.expand_dims(l2, axis)[0][0]


def float_eq(a, b):
    """
    Test if two floats are almost equal to each other
    """
    return abs(a - b) < 1e-10 * max(1e-5, abs(a), abs(b))


def linear_interpol(x, x1, x2, y1, y2):
    """
    Linearly interpolate to point x, between
    the points (x1,y1), (x2,y2)
    """
    assert(x1 <= x)
    assert(x2 >= x)
    assert(x1 <= x2)

    alpha = (x - x1) / (x2 - x1)
    return (1. - alpha) * y1 + alpha * y2


def bracket(fn, x0, dx, args=(), ratio=1.618, maxiter=100):
    """
    Given a function and a starting guess, find two
    inputs for the function that bracket a root.

    Parameters
    ----------
    fn : function
        The function to bracket
    x0 : float
        The starting guess
    dx : float
        Small step for starting the search
    args : parameter list
        Additional arguments to give to fn
    ratio :
        The step size increases by this ratio
        every step in the search. Defaults to
        the golden ratio.
    maxiter : int
        The maximum number of steps before giving up.

    Returns
    -------
    xa, xb, fa, fb: floats
        xa and xb are the inputs which bracket a root of fn.
        fa and fb are the values of the function at those points.
        If the bracket function takes more than maxiter steps,
        it raises a ValueError.
    """
    niter = 0
    dx = np.abs(dx)
    assert(ratio > 1.0)

    # Get the starting positions
    f0 = fn(x0, *args)
    x_left = x0 - dx
    x_right = x0 + dx
    f_left = fn(x_left, *args)
    f_right = fn(x_right, *args)

    # Overshot zero, try making dx smaller
    if (f0 - f_left) * (f_right - f0) < 0.:
        while (f0 - f_left) * (f_right - f0) < 0. and dx > np.finfo('float').eps and niter < maxiter:
            dx /= ratio
            x_left = x0 - dx
            x_right = x0 + dx
            f_left = fn(x_left, *args)
            f_right = fn(x_right, *args)
            niter += 1
        if niter == maxiter:  # Couldn't find something with same slope in both directions
            raise ValueError('Cannot find zero.')

    niter = 0
    slope = f_right - f0
    if slope > 0. and f0 > 0.:  # Walk left
        dx = -dx
        x1 = x_left
        f1 = f_left
    elif slope > 0. and f0 < 0.:  # Walk right
        x1 = x_right
        f1 = f_right
    elif slope < 0. and f0 > 0:  # Walk right
        x1 = x_right
        f1 = f_right
    else:  # Walk left
        dx = -dx
        x1 = x_left
        f1 = f_left

    # Do the walking
    while f0 * f1 > 0. and niter < maxiter:
        dx *= ratio
        xnew = x1 + dx
        fnew = fn(xnew, *args)
        x0 = x1
        f0 = f1
        x1 = xnew
        f1 = fnew
        niter += 1

    if f0 * f1 > 0.:
        raise ValueError('Cannot find zero.')
    else:
        return x0, x1, f0, f1


def _pad_ndarray_inverse_mirror(array, padding):
    """
    Pads an ndarray according to an inverse mirror
    scheme. For example, for a 1D array
    [2, 4, 6, 7, 8] padded by 3 cells, we have:

     padding  |  original array |  padding

    -3 -2  0  |  2  4  6  7  8  |  9 10 12

    Parameters
    ----------
    array : numpy ndarray
        The array to be padded
    padding : tuple
        The number of elements with which to pad the
        array in each dimension.

    Returns
    -------
    padded_array: numpy ndarray
        The padded array

    """
    padded_shape = [n + 2*padding[i] for i, n in enumerate(array.shape)]
    padded_array = np.zeros(padded_shape)

    slices = tuple([slice(padding[i], padding[i] + l) for i, l in enumerate(array.shape)])
    padded_array[slices] = array

    padded_array_indices = list(itertools.product(*[range(n + 2*padding[i]) for i, n in enumerate(array.shape)]))
    inserted_indices = list(itertools.product(*[range(padding[i], padding[i] + l) for i, l in enumerate(array.shape)]))
    padded_array_indices.extend(inserted_indices)

    counter = Counter(padded_array_indices)
    keys = list(counter.keys())
    padded_indices = [keys[i] for i, value in enumerate(counter.values()) if value == 1]
    edge_indices = tuple([tuple([np.min([np.max([axis_idx, padding[dimension]]), padded_array.shape[dimension] - padding[dimension] - 1])
                                 for dimension, axis_idx in enumerate(idx)]) for idx in padded_indices])
    mirror_indices = tuple([tuple([2*edge_indices[i][j] - padded_indices[i][j] for j in range(len(array.shape))]) for i in range(len(padded_indices))])

    for i, idx in enumerate(padded_indices):
        padded_array[idx] = 2.*padded_array[edge_indices[i]] - padded_array[mirror_indices[i]]

    return padded_array


def smooth_array(array, grid_spacing,
                 gaussian_rms_widths, truncate=4.0,
                 mode='inverse_mirror'):
    """
    Creates a smoothed array by convolving it with a gaussian filter.
    Grid resolutions and gaussian RMS widths are required for each of
    the axes of the numpy array. The smoothing is truncated at a
    user-defined number of standard deviations. The edges of the array
    can be padded in a number of different ways given by the
    'mode' parameter.

    Parameters
    ----------
    array : numpy ndarray
        The array to smooth
    grid_spacing : numpy array of floats
        The spacing of points along each axis
    gaussian_rms_widths : numpy array of floats
        The Gaussian RMS widths/standard deviations for the
        Gaussian convolution.
    truncate : float (default=4.)
        The number of standard deviations at which to truncate
        the smoothing.
    mode : {'reflect', 'constant', 'nearest', 'mirror', 'wrap', 'inverse_mirror'}
        The mode parameter determines how the array borders are handled
        either by scipy.ndimage.filters.gaussian_filter.
        Default is 'inverse_mirror', which uses
        :func:`burnman.tools.math._pad_ndarray_inverse_mirror`.

    Returns
    -------
    smoothed_array: numpy ndarray
       The smoothed array

    """

    # gaussian_filter works with standard deviations normalised to
    # the grid spacing.
    sigma = tuple(np.array(gaussian_rms_widths)/np.array(grid_spacing))

    if mode == 'inverse_mirror':
        padding = tuple([int(np.ceil(truncate*s)) for s in sigma])
        padded_array = _pad_ndarray_inverse_mirror(array, padding)
        smoothed_padded_array = gaussian_filter(padded_array,
                                                sigma=sigma)
        slices = tuple([slice(padding[i], padding[i] + l) for i, l in enumerate(array.shape)])
        smoothed_array = smoothed_padded_array[slices]
    else:
        smoothed_array = gaussian_filter(array, sigma=sigma, mode=mode)

    return smoothed_array


def interp_smoothed_array_and_derivatives(array,
                                          x_values, y_values,
                                          x_stdev=0., y_stdev=0.,
                                          truncate=4.,
                                          mode='inverse_mirror',
                                          indexing='xy'):
    """
    Creates a smoothed array on a regular 2D grid. Smoothing
    is achieved using :func:`burnman.tools.math.smooth_array`.
    Outputs scipy.interpolate.interp2d() interpolators
    which can be used to query the array, or its derivatives in the
    x- and y- directions.

    Parameters
    ----------
    array : 2D numpy array
        The array to smooth. Each element array[i][j]
        corresponds to the position x_values[i], y_values[j]
    x_values : 1D numpy array
        The gridded x values over which to create the smoothed grid
    y_values : 1D numpy array
        The gridded y_values over which to create the smoothed grid
    x_stdev : float
        The standard deviation for the Gaussian filter along the x axis
    y_stdev : float
        The standard deviation for the Gaussian filter along the x axis
    truncate : float (optional)
        The number of standard deviations at which to truncate
        the smoothing (default = 4.).
    mode : {'reflect', 'constant', 'nearest', 'mirror', 'wrap', 'inverse_mirror'}
        The mode parameter determines how the array borders are handled
        either by scipy.ndimage.filters.gaussian_filter.
        Default is 'inverse_mirror', which uses
        :func:`burnman.tools.math._pad_ndarray_inverse_mirror`.
    indexing : {'xy', 'ij'}, optional
        Cartesian ('xy', default) or matrix ('ij') indexing of output.
        See numpy.meshgrid for more details.

    Returns
    -------
    interps: tuple of three interp2d functors
        interpolation functions for the smoothed property and
        the first derivatives with respect to x and y.

    """

    dx = x_values[1] - x_values[0]
    dy = y_values[1] - y_values[0]

    if indexing == 'xy':
        smoothed_array = smooth_array(array=array,
                                      grid_spacing=np.array([dy, dx]),
                                      gaussian_rms_widths=np.array([y_stdev, x_stdev]),
                                      truncate=truncate,
                                      mode=mode)

    elif indexing == 'ij':
        smoothed_array = smooth_array(array=array,
                                      grid_spacing=np.array([dx, dy]),
                                      gaussian_rms_widths=np.array([x_stdev, y_stdev]),
                                      truncate=truncate,
                                      mode=mode).T

    else:
        raise Exception('Indexing scheme not recognised. Should be ij or xy.')

    dSAdydy, dSAdxdx = np.gradient(smoothed_array)

    interps = (interp2d(x_values, y_values, smoothed_array, kind='linear'),
               interp2d(x_values, y_values, dSAdxdx/dx, kind='linear'),
               interp2d(x_values, y_values, dSAdydy/dy, kind='linear'))

    return interps


def compare_l2(depth, calc, obs):
    """
    Computes the L2 norm for N profiles at a time (assumed to be linear between points).

    :type depths: array of float
    :param depths: depths. :math:`[m]`
    :type calc: list of arrays of float
    :param calc: N arrays calculated values, e.g. [mat_vs,mat_vphi]
    :type obs: list of arrays of float
    :param obs: N arrays of values (observed or calculated) to compare to , e.g. [seis_vs, seis_vphi]

    :returns: array of L2 norms of length N
    :rtype: array of floats
    """
    err = []
    for i in range(len(calc)):
        err.append(l2(depth, calc[i], obs[i]))

    return err


def compare_chifactor(calc, obs):
    """
    Computes the chi factor for N profiles at a time. Assumes a 1% a priori uncertainty on the seismic model.


    :type calc: list of arrays of float
    :param calc: N arrays calculated values, e.g. [mat_vs,mat_vphi]
    :type obs: list of arrays of float
    :param obs: N arrays of values (observed or calculated) to compare to , e.g. [seis_vs, seis_vphi]

    :returns: error array of length N
    :rtype: array of floats
    """
    err = []
    for i in range(len(calc)):
        err.append(chi_factor(calc[i], obs[i]))

    return err


def l2(x, funca, funcb):
    """
    Computes the L2 norm for one profile(assumed to be linear between points).

    :type x: array of float
    :param x: depths :math:`[m]`.
    :type funca: list of arrays of float
    :param funca: array calculated values
    :type funcb: list of arrays of float
    :param funcb: array of values (observed or calculated) to compare to

    :returns: L2 norm
    :rtype: array of floats
    """
    diff = np.array(funca - funcb)
    diff = diff * diff

    return integrate.trapz(diff, x)


def nrmse(x, funca, funcb):
    """
    Normalized root mean square error for one profile
    :type x: array of float
    :param x: depths in m.
    :type funca: list of arrays of float
    :param funca: array calculated values
    :type funcb: list of arrays of float
    :param funcb: array of values (observed or calculated) to compare to

    :returns: RMS error
    :rtype: array of floats

    """
    diff = np.array(funca - funcb)
    diff = diff * diff
    rmse = np.sqrt(np.sum(diff) / x)
    nrmse = rmse / (np.max(funca) - np.min(funca))

    return nrmse


def chi_factor(calc, obs):
    """
    :math:`\\chi` factor for one profile assuming 1% uncertainty on the reference model (obs)
    :type calc: list of arrays of float
    :param calc: array calculated values
    :type obs: list of arrays of float
    :param obs: array of reference values to compare to

    :returns: :math:`\\chi` factor
    :rtype: array of floats

    """

    err = np.empty_like(calc)
    for i in range(len(calc)):
        err[i] = pow((calc[i] - obs[i]) / (0.01 * np.mean(obs)), 2.)

    err_tot = np.sum(err) / len(err)

    return err_tot


def independent_row_indices(array):
    """
    Returns the indices corresponding to an independent set of rows
    for a given array. The independent rows are determined from the pivots
    used during row reduction/Gaussian elimination.

    Parameters
    ----------
    array : 2D numpy array of floats
        The input array

    Returns
    -------
    indices : 1D numpy array of integers
        The indices corresponding to a set of independent rows
        of the input array.
    """
    m = Matrix(array.shape[0], array.shape[1],
               lambda i, j: Rational(array[i, j]).limit_denominator(1000))
    _, pivots, swaps = row_reduce(m, iszerofunc=lambda x: x.is_zero,
                                  simpfunc=lambda x: Rational(x).limit_denominator(1000))
    indices = np.array(range(len(array)))
    for swap in np.array(swaps):
        indices[swap] = indices[swap[::-1]]
    return sorted(indices[:len(pivots)])


def generate_complete_basis(incomplete_basis, array):
    """
    Given a 2D array with independent rows and a second 2D array that spans a
    larger space, creates a complete basis for the combined array using all
    the rows of the first array, followed by any required rows of the
    second array. So, for example, if the first array is:
    [[1, 0, 0], [1, 1, 0]] and the second array is:
    [[1, 0, 0], [0, 1, 0], [0, 0, 1]], the complete basis will be:
    [[1, 0, 0], [1, 1, 0], [0, 0, 1]].

    Parameters
    ----------
    incomplete_basis : 2D numpy array
        An array containing the basis to be completed.

    array : 2D numpy array
        An array spanning the full space for which a basis is required.

    Returns
    -------
    complete_basis : 2D numpy array
        An array containing the basis vectors spanning both of the
        input arrays.
    """

    a = np.concatenate((incomplete_basis, array))
    return a[independent_row_indices(a)]



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
Equation of State
-----------------


"""
from __future__ import absolute_import

from .equation_of_state import EquationOfState
from .murnaghan import Murnaghan
from .birch_murnaghan import BM2, BM3
from .birch_murnaghan_4th import BM4
from .mie_grueneisen_debye import MGD2, MGD3
from .slb import SLB2, SLB3
from .modified_tait import MT
from .hp import HP_TMT
from .hp import HP_TMTL
from .hp import HP98
from .cork import CORK
from .vinet import Vinet
from .morse_potential import Morse
from .reciprocal_kprime import RKprime
from .dks_liquid import DKS_L
from .dks_solid import DKS_S
from .aa import AA

from .property_modifiers import calculate_property_modifications

from .helper import create



from __future__ import absolute_import
# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


import scipy.optimize as opt
from scipy.special import gamma, gammainc
from . import equation_of_state as eos
from ..tools.math import bracket
import warnings
import numpy as np

# Try to import the jit from numba.  If it is
# not available, just go with the standard
# python interpreter
try:
    from numba import jit
except ImportError:
    def jit(fn):
        return fn


@jit
def _delta_PoverK_from_P(PoverK, pressure, K_0, Kprime_0, Kprime_inf):
    return PoverK - (pressure/K_0)*np.power((1. - Kprime_inf*PoverK), Kprime_0/Kprime_inf) # eq. 58

@jit
def _delta_PoverK_from_V(PoverK, V, V_0, K_0, Kprime_0, Kprime_inf):
    Kprime_ratio = Kprime_0 / Kprime_inf
    return ( np.log( V_0 / V ) +
             Kprime_ratio / Kprime_inf * np.log(1. - Kprime_inf * PoverK) +
             (Kprime_ratio - 1.) * PoverK ) # eq. 61

def _upper_incomplete_gamma(z, a):
    """
    An implementation of the non-regularised upper incomplete gamma
    function. Computed using the relationship with the regularised
    lower incomplete gamma function (scipy.special.gammainc).
    Uses the recurrence relation wherever z<0.
    """
    n = int(-np.floor(z))
    if n > 0:
        z = z + n
        u_gamma = (1. - gammainc(z, a))*gamma(z)

        for i in range(n):
            z = z - 1.
            u_gamma = (u_gamma - np.power(a, z)*np.exp(-a))/z
        return u_gamma
    else:
        return (1. - gammainc(z, a))*gamma(z)

def _PoverK_from_P(pressure, params):
    """
    Calculates the pressure:bulk modulus ratio
    from a given pressure using brentq optimization
    """
    args = ((pressure - params['P_0']), params['K_0'],
            params['Kprime_0'], params['Kprime_inf'])
    return opt.brentq(_delta_PoverK_from_P,
                      1./(params['Kprime_inf'] - params['Kprime_0']) + np.finfo(float).eps,
                      1./params['Kprime_inf'] - np.finfo(float).eps,
                      args=args)

def _PoverK_from_V(volume, params):
    """
    Calculates the pressure:bulk modulus ratio
    from a given volume using brentq optimization
    """
    args = (volume, params['V_0'], params['K_0'],
            params['Kprime_0'], params['Kprime_inf'])
    return opt.brentq(_delta_PoverK_from_V,
                      1./(params['Kprime_inf'] - params['Kprime_0']) + np.finfo(float).eps,
                      1./params['Kprime_inf'] - np.finfo(float).eps,
                      args=args)

def bulk_modulus(pressure, params):
    """
    Returns the bulk modulus at a given pressure
    """
    PoverK = _PoverK_from_P(pressure, params)
    K = params['K_0']*np.power((1. - params['Kprime_inf']*PoverK), -
                               params['Kprime_0']/params['Kprime_inf'])
    return K

def shear_modulus(pressure, params):
    """
    Returns the shear modulus at a given pressure
    """
    G = ( params['G_0']/params['K_0'] * bulk_modulus(pressure, params) -
          (params['G_0']/params['K_0']*params['Kprime_inf'] - params['Gprime_inf']) * pressure )
    return G # eq. 78


class RKprime(eos.EquationOfState):

    """
    Class for the isothermal reciprocal K-prime equation of state
    detailed in :cite:`StaceyDavis2004`.  This equation of state is
    a development of work by :cite:`Keane1954` and :cite:`Stacey2000`,
    making use of the fact that :math:`K'` typically varies smoothly
    as a function of :math:`P/K`, and is thermodynamically required to
    exceed 5/3 at infinite pressure.

    It is worth noting that this equation of state rapidly becomes
    unstable at negative pressures, so should not be trusted to provide
    a good *HT-LP* equation of state using a thermal pressure
    formulation. The negative root of :math:`dP/dK`
    can be found at :math:`K/P = K'_{\infty} - K'_0`,
    which corresponds to a bulk modulus of
    :math:`K = K_0 ( 1 - K'_{\infty}/K'_0 )^{K'_0/K'_{\infty}}`
    and a volume of
    :math:`V = V_0 ( K'_0 / (K'_0 - K'_{\infty}) )^{K'_0/{K'}^2_{\infty}} \exp{(-1/K'_{\infty})}`.

    This equation of state has no temperature dependence.
    """

    def volume(self, pressure, temperature, params):
        """
        Returns volume :math:`[m^3]` as a function of pressure :math:`[Pa]`.
        """
        Kprime_ratio = params['Kprime_0']/params['Kprime_inf']
        PoverK = _PoverK_from_P(pressure, params)

        V = params['V_0'] * np.exp( Kprime_ratio/params['Kprime_inf'] *
                                    np.log(1. - params['Kprime_inf'] * PoverK) +
                                    (Kprime_ratio - 1.) * PoverK ) # Eq. 61

        return V

    def pressure(self, temperature, volume, params):
        """
        Returns pressure :math:`[Pa]` as a function of volume :math:`[m^3]`.
        """
        PoverK = _PoverK_from_V(volume, params)
        return params['P_0'] + ( params['K_0'] * PoverK *
                                 np.power(1. - params['Kprime_inf'] * PoverK,
                                          -params['Kprime_0']/params['Kprime_inf']) )

    def isothermal_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus :math:`K_T` :math:`[Pa]` as a function of pressure :math:`[Pa]`,
        temperature :math:`[K]` and volume :math:`[m^3]`.
        """
        return bulk_modulus(pressure, params)

    def adiabatic_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus :math:`K_s` of the mineral. :math:`[Pa]`.
        """
        return bulk_modulus(pressure, params)

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Returns shear modulus :math:`G` of the mineral. :math:`[Pa]`
        """
        return shear_modulus(pressure, params)

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the molar entropy :math:`\mathcal{S}` of the mineral. :math:`[J/K/mol]`
        """
        return 0.

    def _intVdP(self, xi, params):

        a = params['Kprime_inf']
        b = (params['Kprime_0']/params['Kprime_inf']/params['Kprime_inf'] -
             params['Kprime_0']/params['Kprime_inf'] - 1.)
        c = params['Kprime_0'] - params['Kprime_inf']
        f = (params['Kprime_0']/params['Kprime_inf'] - 1.)

        i1 = float( params['V_0'] * params['K_0'] *
                    np.exp(f / a) * np.power(a, b - 1.) /
                    np.power(f, b + 2.) *
                    ( f * params['Kprime_0'] * _upper_incomplete_gamma( b + 1. ,
                                                                        f * (1./a - xi) ) -
                      a * c * _upper_incomplete_gamma( b + 2., f * (1./a - xi) ) ) )

        return i1

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Gibbs free energy :math:`\mathcal{G}` of the mineral. :math:`[J/mol]`
        """
        # G = E0 + int VdP (when S = 0)
        K = self.isothermal_bulk_modulus(pressure, temperature, volume, params)
        return params['E_0'] + params['P_0']*params['V_0'] + self._intVdP((pressure - params['P_0'])/K, params) - self._intVdP(0., params)

    def molar_internal_energy(self, pressure, temperature, volume, params):
        """
        Returns the internal energy :math:`\mathcal{E}` of the mineral. :math:`[J/mol]`
        """
        # E = G - PV (+ TS)
        return ( self.gibbs_free_energy(pressure, temperature, volume, params) - pressure*volume)

    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return a very large number. :math:`[J/K/mol]`
        """
        return 1.e99

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return a very large number. :math:`[J/K/mol]`
        """
        return 1.e99

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return zero. :math:`[1/K]`
        """
        return 0.

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return zero. :math:`[unitless]`
        """
        return 0.

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters.
        The value for :math:`K'_{\infty}` is thermodynamically bounded
        between 5/3 and :math:`K'_0` :cite:`StaceyDavis2004`.
        """

        if 'E_0' not in params:
            params['E_0'] = 0.
        if 'P_0' not in params:
            params['P_0'] = 0.

        # If G and Gprime_inf are not included this is presumably deliberate,
        # as we can model density and bulk modulus just fine without them,
        # so just add them to the dictionary as nans
        if 'G_0' not in params:
            params['G_0'] = float('nan')
        if 'Gprime_inf' not in params:
            params['Gprime_inf'] = float('nan')

        # Check that all the required keys are in the dictionary
        expected_keys = ['V_0', 'K_0', 'Kprime_0', 'Kprime_inf', 'G_0', 'Gprime_inf']
        for k in expected_keys:
            if k not in params:
                raise KeyError('params object missing parameter : ' + k)

        # Finally, check that the values are reasonable.
        if params['P_0'] < 0.:
            warnings.warn('Unusual value for P_0', stacklevel=2)
        if params['V_0'] < 1.e-7 or params['V_0'] > 1.e-3:
            warnings.warn('Unusual value for V_0', stacklevel=2)
        if params['K_0'] < 1.e9 or params['K_0'] > 1.e13:
            warnings.warn('Unusual value for K_0', stacklevel=2)
        if params['Kprime_0'] < 0. or params['Kprime_0'] > 10.:
            warnings.warn('Unusual value for Kprime_0', stacklevel=2)
        if params['Kprime_inf'] < 5./3. or params['Kprime_inf'] > params['Kprime_0']:
            warnings.warn('Unusual value for Kprime_inf', stacklevel=2) # eq. 17
        if params['G_0'] < 0.0 or params['G_0'] > 1.e13:
            warnings.warn('Unusual value for G_0', stacklevel=2)
        if params['Gprime_inf'] < -5. or params['Gprime_inf'] > 10.:
            warnings.warn('Unusual value for Gprime_inf', stacklevel=2)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import
from __future__ import print_function

from subprocess import Popen, PIPE, STDOUT
from os import rename

import numpy as np
from scipy.interpolate import interp2d, griddata

from .material import Material, material_property

from ..tools.misc import copy_documentation


def create_perplex_table(werami_path, project_name, outfile, n_pressures, n_temperatures, pressure_range=None, temperature_range=None):
    '''
    This function uses PerpleX's werami software to output a table file containing the following material properties.
    2 - Density (kg/m3)
    4 - Expansivity (1/K, for volume)
    5 - Compressibility (1/bar, for volume)
    10 - Adiabatic bulk modulus (bar)
    11 - Adiabatic shear modulus (bar)
    12 - Sound velocity (km/s)
    13 - P-wave velocity (Vp, km/s)
    14 - S-wave velocity (Vs, km/s)
    17 - Entropy (J/K/kg)
    18 - Enthalpy (J/kg)
    19 - Heat Capacity (J/K/kg)
    22 - Molar Volume (J/bar)
    '''

    print('Working on creating {0}x{1} P-T table file using werami. Please wait.\n'.format(
        n_pressures, n_temperatures))

    try:
        str2 = 'y\n{0} {1}\n{2} {3}\n'.format(pressure_range[0]/1.e5, pressure_range[1]/1.e5,
                                              temperature_range[0], temperature_range[1])
    except:
        print('Keeping P-T range the same as the original project range.\n')
        str2 = 'n\n'

    stdin = '{0:s}\n2\n' \
        '2\nn\n' \
        '4\nn\n' \
        '5\nn\n' \
        '10\nn\n' \
        '11\nn\n' \
        '12\nn\n' \
        '13\nn\n' \
        '14\nn\n' \
        '17\nn\n' \
        '18\nn\n' \
        '19\nn\n' \
        '22\nn\n' \
        '0\n' \
        '{1:s}' \
        '{2:d} {3:d}\n' \
        '0'.format(project_name, str2, n_pressures, n_temperatures)

    p = Popen(werami_path, stdout=PIPE, stdin=PIPE,
              stderr=STDOUT, encoding='utf8')
    stdout = p.communicate(input=stdin)[0]
    print(stdout)
    out = [s for s in stdout.split(
        '\n') if "Output has been written to the" in s][0].split()[-1]
    rename(out, outfile)
    print('Output file renamed to {0:s}'.format(outfile))
    print('Processing complete')


class PerplexMaterial(Material):
    """
    This is the base class for a PerpleX material. States of the material
    can only be queried after setting the pressure and temperature
    using set_state().

    Instances of this class are initialised with
    a 2D PerpleX tab file. This file should be in the standard format
    (as output by werami), and should have columns with the following names:
    'rho,kg/m3', 'alpha,1/K', 'beta,1/bar', 'Ks,bar', 'Gs,bar', 'v0,km/s',
    'vp,km/s', 'vs,km/s', 's,J/K/kg', 'h,J/kg', 'cp,J/K/kg', 'V,J/bar/mol'.
    The order of these names is not important.

    Properties of the material are determined by linear interpolation from
    the PerpleX grid. They are all returned in SI units on a molar basis,
    even though the PerpleX tab file is not in these units.

    This class is available as ``burnman.PerplexMaterial``.
    """

    def __init__(self, tab_file, name='Perple_X material'):
        self.name = name
        self.params = {'name': name}
        self._property_interpolators, self.params['molar_mass'], self.bounds = self._read_2D_perplex_file(
            tab_file)
        Material.__init__(self)

    def _read_2D_perplex_file(self, filename):
        with open(filename, 'r') as f:
            datastream = f.read()

        lines = [line.strip().split()
                 for line in datastream.split('\n') if line.strip()]

        if lines[2][0] != '2':
            raise Exception('This is not a 2D PerpleX table')

        bounds = [(float(lines[4][0]), float(lines[5][0]), int(lines[6][0])),
                  (float(lines[8][0]), float(lines[9][0]), int(lines[10][0]))]

        if lines[3][0] == 'P(bar)' and lines[7][0] == 'T(K)':
            Pmin, Pint, nP = bounds[0]
            Tmin, Tint, nT = bounds[1]
        elif lines[3][0] == 'T(K)' and lines[7][0] == 'P(bar)':
            Pmin, Pint, nP = bounds[1]
            Tmin, Tint, nT = bounds[0]
        else:
            raise Exception('This file does not have a recognised PerpleX structure.\n'
                            'Are the independent variables P(bar) and T(K)?')

        Pmin = Pmin*1.e5  # bar to Pa
        Pint = Pint*1.e5  # bar to Pa
        Pmax = Pmin + Pint*(nP-1.)
        Tmax = Tmin + Tint*(nT-1.)
        pressures = np.linspace(Pmin, Pmax, nP)
        temperatures = np.linspace(Tmin, Tmax, nT)
        n_properties = int(lines[11][0])
        property_list = lines[12]

        # property_table[i][j][k] returns the kth property at the ith pressure and jth temperature
        table = np.array([[float(string) for string in line]
                         for line in lines[13:13+nP*nT]])

        if lines[3][0] == 'P(bar)':
            property_table = np.swapaxes(
                table.reshape(nT, nP, n_properties), 0, 1)
        else:
            property_table = table.reshape(nP, nT, n_properties)

        ordered_property_list = ['rho,kg/m3',
                                 'alpha,1/K',
                                 'beta,1/bar',
                                 'Ks,bar',
                                 'Gs,bar',
                                 'v0,km/s',
                                 'vp,km/s',
                                 'vs,km/s',
                                 's,J/K/kg',
                                 'h,J/kg',
                                 'cp,J/K/kg',
                                 'V,J/bar/mol']
        p_indices = [i for i, p in enumerate(
            property_list) for ordered_p in ordered_property_list if p == ordered_p]

        properties = {}
        for i, p_idx in enumerate(p_indices):
            # Fill in NaNs as long as they aren't in the corners of the P-T grid
            a = np.array(property_table[:, :, [p_idx]][:, :, 0])
            x, y = np.indices(a.shape)
            a[np.isnan(a)] = griddata((x[~np.isnan(a)], y[~np.isnan(a)]),  # points we know
                                      a[~np.isnan(a)],  # values we know
                                      (x[np.isnan(a)], y[np.isnan(a)]))

            # Fill any remaining NaNs with zeros
            properties[ordered_property_list[i]] = np.nan_to_num(a, 0.)

        densities = properties['rho,kg/m3']
        volumes = 1.e-5 * properties['V,J/bar/mol']
        molar_masses = densities*volumes
        molar_mass = np.mean(molar_masses)
        property_interpolators = {'rho': interp2d(pressures, temperatures, densities.T, bounds_error=True),
                                  'alpha': interp2d(pressures, temperatures, properties['alpha,1/K'].T, bounds_error=True),
                                  'K_T': interp2d(pressures, temperatures, 1.e5 / properties['beta,1/bar'].T, bounds_error=True),
                                  'K_S': interp2d(pressures, temperatures, 1.e5 * properties['Ks,bar'].T, bounds_error=True),
                                  'G_S': interp2d(pressures, temperatures, 1.e5 * properties['Gs,bar'].T, bounds_error=True),
                                  'bulk_sound_velocity': interp2d(pressures, temperatures, 1.e3*properties['v0,km/s'].T, bounds_error=True),
                                  'p_wave_velocity': interp2d(pressures, temperatures, 1.e3*properties['vp,km/s'].T, bounds_error=True),
                                  's_wave_velocity': interp2d(pressures, temperatures, 1.e3*properties['vs,km/s'].T, bounds_error=True),
                                  'S': interp2d(pressures, temperatures, properties['s,J/K/kg'].T*molar_masses.T, bounds_error=True),
                                  'H': interp2d(pressures, temperatures, properties['h,J/kg'].T*molar_masses.T, bounds_error=True),
                                  'C_p': interp2d(pressures, temperatures, properties['cp,J/K/kg'].T*molar_masses.T, bounds_error=True),
                                  'V': interp2d(pressures, temperatures, volumes.T, bounds_error=True)}

        bounds = [[Pmin, Pmax], [Tmin, Tmax]]
        return property_interpolators, molar_mass, bounds

    @copy_documentation(Material.set_state)
    def set_state(self, pressure, temperature):

        if not np.logical_and(np.all(self.bounds[0][0] <= pressure),
                              np.all(pressure <= self.bounds[0][1])):
            raise ValueError("The set_state pressure ({0:.4f}) is outside the bounds of this rock ({1:.4f}-{2:.4f} GPa)".format(pressure,
                                                                                                                                self.bounds[
                                                                                                                                    0][0]/1.e9,
                                                                                                                                self.bounds[0][1]/1.e9))
        if not np.logical_and(np.all(self.bounds[1][0] <= temperature),
                              np.all(temperature <= self.bounds[1][1])):
            raise ValueError("The set_state temperature ({0:.1f}) is outside the bounds of this rock ({1:.1f}-{2:.1f} K)".format(temperature,
                                                                                                                                 self.bounds[
                                                                                                                                     1][0],
                                                                                                                                 self.bounds[1][1]))
        Material.set_state(self, pressure, temperature)

    """
    Properties by linear interpolation of Perple_X output
    """

    @material_property
    @copy_documentation(Material.molar_volume)
    def molar_volume(self):
        return self._property_interpolators['V'](self.pressure, self.temperature)[0]

    @material_property
    @copy_documentation(Material.molar_enthalpy)
    def molar_enthalpy(self):
        return self._property_interpolators['H'](self.pressure, self.temperature)[0]

    @material_property
    @copy_documentation(Material.molar_entropy)
    def molar_entropy(self):
        return self._property_interpolators['S'](self.pressure, self.temperature)[0]

    @material_property
    @copy_documentation(Material.isothermal_bulk_modulus)
    def isothermal_bulk_modulus(self):
        return self._property_interpolators['K_T'](self.pressure, self.temperature)[0]

    @material_property
    @copy_documentation(Material.adiabatic_bulk_modulus)
    def adiabatic_bulk_modulus(self):
        return self._property_interpolators['K_S'](self.pressure, self.temperature)[0]

    @material_property
    @copy_documentation(Material.molar_heat_capacity_p)
    def molar_heat_capacity_p(self):
        return self._property_interpolators['C_p'](self.pressure, self.temperature)[0]

    @material_property
    @copy_documentation(Material.thermal_expansivity)
    def thermal_expansivity(self):
        return self._property_interpolators['alpha'](self.pressure, self.temperature)[0]

    @material_property
    @copy_documentation(Material.shear_modulus)
    def shear_modulus(self):
        return self._property_interpolators['G_S'](self.pressure, self.temperature)[0]

    @material_property
    @copy_documentation(Material.p_wave_velocity)
    def p_wave_velocity(self):
        return self._property_interpolators['p_wave_velocity'](self.pressure, self.temperature)[0]

    @material_property
    @copy_documentation(Material.bulk_sound_velocity)
    def bulk_sound_velocity(self):
        return self._property_interpolators['bulk_sound_velocity'](self.pressure, self.temperature)[0]

    @material_property
    @copy_documentation(Material.shear_wave_velocity)
    def shear_wave_velocity(self):
        return self._property_interpolators['s_wave_velocity'](self.pressure, self.temperature)[0]

    """
    Properties from mineral parameters,
    Legendre transformations
    or Maxwell relations
    """

    @material_property
    @copy_documentation(Material.molar_gibbs)
    def molar_gibbs(self):
        return self.molar_enthalpy - self.temperature*self.molar_entropy

    @material_property
    @copy_documentation(Material.molar_mass)
    def molar_mass(self):
        if 'molar_mass' in self.params:
            return self.params['molar_mass']
        else:
            raise ValueError(
                "No molar_mass parameter for mineral " + self.to_string + ".")

    @material_property
    @copy_documentation(Material.density)
    def density(self):
        return self._property_interpolators['rho'](self.pressure, self.temperature)[0]

    @material_property
    @copy_documentation(Material.molar_internal_energy)
    def molar_internal_energy(self):
        return self.molar_gibbs - self.pressure * self.molar_volume + self.temperature * self.molar_entropy

    @material_property
    @copy_documentation(Material.molar_helmholtz)
    def molar_helmholtz(self):
        return self.molar_gibbs - self.pressure * self.molar_volume

    @material_property
    @copy_documentation(Material.isothermal_compressibility)
    def isothermal_compressibility(self):
        return 1. / self.isothermal_bulk_modulus

    @material_property
    @copy_documentation(Material.adiabatic_compressibility)
    def adiabatic_compressibility(self):
        return 1. / self.adiabatic_bulk_modulus

    @material_property
    @copy_documentation(Material.molar_heat_capacity_v)
    def molar_heat_capacity_v(self):
        return (self.molar_heat_capacity_p
                - self.molar_volume * self.temperature
                * self.thermal_expansivity * self.thermal_expansivity
                * self.isothermal_bulk_modulus)

    @material_property
    @copy_documentation(Material.grueneisen_parameter)
    def grueneisen_parameter(self):
        return (self.thermal_expansivity
                * self.molar_volume
                * self.adiabatic_bulk_modulus
                / self.molar_heat_capacity_p)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
Mineral database

  - :mod:`~burnman.minerals.SLB_2005`
  - :mod:`~burnman.minerals.SLB_2011_ZSB_2013`
  - :mod:`~burnman.minerals.SLB_2011`
  - :mod:`~burnman.minerals.DKS_2013_liquids`
  - :mod:`~burnman.minerals.DKS_2013_solids`
  - :mod:`~burnman.minerals.RS_2014_liquids`
  - :mod:`~burnman.minerals.Murakami_etal_2012`
  - :mod:`~burnman.minerals.Murakami_2013`
  - :mod:`~burnman.minerals.Matas_etal_2007`
  - :mod:`~burnman.minerals.HP_2011_ds62`
  - :mod:`~burnman.minerals.HP_2011_fluids`
  - :mod:`~burnman.minerals.HHPH_2013`
  - :mod:`~burnman.minerals.HGP_2018_ds633`
  - :mod:`~burnman.minerals.other`
"""
from __future__ import absolute_import

# Stixrude and Lithgow-Bertelloni
from . import SLB_2011
from . import SLB_2011_ZSB_2013
from . import SLB_2005

# ab initio
from . import RS_2014_liquids
from . import DKS_2013_liquids
from . import DKS_2013_solids

# Murakami and coworkers
from . import Murakami_etal_2012
from . import Murakami_2013

# Matas and coworkers
from . import Matas_etal_2007

# Holland, Powell and coworkers
from . import HP_2011_ds62
from . import HP_2011_fluids
from . import HHPH_2013
from . import JH_2015
from . import HGP_2018_ds633

# Kurnosov et al. 2017
from . import KMFBZ_2017

# Irving et al. 2018
from . import ICL_2018

# Other
from . import Sundman_1991
from . import other



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import
from __future__ import print_function

import operator
import bisect
import pkgutil
from collections import Counter, OrderedDict
import numpy as np
from .math import round_to_n, linear_interpol


class OrderedCounter(Counter, OrderedDict):
    """
    Counter that remembers the order elements are first encountered
    """

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, OrderedDict(self))

    def __reduce__(self):
        return self.__class__, (OrderedDict(self),)


def copy_documentation(copy_from):
    """
    Decorator @copy_documentation(another_function) will copy the documentation found in a different
    function (for example from a base class). The docstring applied to some function a() will be ::

        (copied from BaseClass.some_function):
        <documentation from BaseClass.some_function>
        <optionally the documentation found in a()>

    """
    def mydecorator(func):
        def wrapper(*args):
            return func(*args)
        old = ""
        if func.__doc__:
            old = "\n" + func.__doc__

        copied_from = ""
        if hasattr(copy_from, "__name__"):
            copied_from = "(copied from " + copy_from.__name__ + "):\n"
        wrapper.__doc__ = copied_from + copy_from.__doc__ + old
        wrapper.__name__ = func.__name__
        return wrapper
    return mydecorator


def merge_two_dicts(x, y):
    """Given two dicts, merge them into a new dict as a shallow copy."""
    z = x.copy()
    z.update(y)
    return z


def flatten(arr):
    return (flatten(arr[0]) + (flatten(arr[1:]) if len(arr) > 1 else [])
            if type(arr) is list or type(arr) is np.ndarray else [arr])


def pretty_print_values(popt, pcov, params):
    """
    Takes a numpy array of parameters, the corresponding covariance matrix
    and a set of parameter names and prints the parameters and
    principal 1-s.d.uncertainties (np.sqrt(pcov[i][i]))
    in a nice text based format.
    """
    for i, p in enumerate(params):
        p_rnd = round_to_n(popt[i], np.sqrt(pcov[i][i]), 1)
        c_rnd = round_to_n(np.sqrt(pcov[i][i]), np.sqrt(pcov[i][i]), 1)

        if p_rnd != 0.:
            p_expnt = np.floor(np.log10(np.abs(p_rnd)))
        else:
            p_expnt = 0.

        scale = np.power(10., p_expnt)
        nd = p_expnt - np.floor(np.log10(np.abs(c_rnd)))
        print('{0:s}: ({1:{4}{5}f} +/- {2:{4}{5}f}) x {3:.0e}'.format(p, p_rnd/scale, c_rnd/scale, scale, 0, (nd)/10.))


def pretty_print_table(table, use_tabs=False):
    """
    Takes a 2d table and prints it in a nice text based format. If
    use_tabs=True then only \t is used as a separator. This is useful for
    importing the data into other apps (Excel, ...). The default is to pad
    the columns with spaces to make them look neat. The first column is
    left aligned, while the remainder is right aligned.
    """
    if use_tabs:
        for r in table:
            print("\t".join(r).replace("_", "\_"))
        return

    def col_width(table, colidx):
        return max([len(str(row[colidx])) for row in table])

    # create a format string with the first column left aligned, the others right
    # example:   {:<27}{:>11}{:>6}{:>8}
    frmt = "".join(
        [('{:<' if i == 0 else '{:>') + str(1 + col_width(table, i)) + '}' for i in range(len(table[0]))])
    for r in table:
        print(frmt.format(*r))


def sort_table(table, col=0):
    """
    Sort the table according to the column number
    """
    return sorted(table, key=operator.itemgetter(col))


def read_table(filename):
    datastream = pkgutil.get_data('burnman', 'data/' + filename)
    datalines = [line.strip()
                 for line in datastream.decode('ascii').split('\n') if line.strip()]
    table = []

    for line in datalines:
        if (line[0] != '#'):
            numbers = np.fromstring(line, sep=' ')
            table.append(numbers)
    return np.array(table)


def array_from_file(filename):
    """
    Generic function to read a file containing floats and commented lines
    into a 2D numpy array.

    Commented lines are prefixed by the characters # or %.
    """
    f = open(filename, 'r')
    data = []
    datastream = f.read()
    f.close()
    datalines = [line.strip().split()
                 for line in datastream.split('\n') if line.strip()]
    for line in datalines:
        if line[0] != "#" and line[0] != "%":
            data.append(map(float, line))

    data = np.array(zip(*data))
    return data


def cut_table(table, min_value, max_value):
    tablen = []
    for i in range(min_value, max_value, 1):
        tablen.append(table[i, :])
    return tablen


def lookup_and_interpolate(table_x, table_y, x_value):
    idx = bisect.bisect_left(table_x, x_value) - 1
    if (idx < 0):
        return table_y[0]
    elif (idx < len(table_x) - 1):
        return linear_interpol(x_value, table_x[idx], table_x[idx + 1],
                               table_y[idx], table_y[idx + 1])
    else:
        return table_y[idx]


def attribute_function(m, attributes, powers=[]):
    """
    Function which returns a function which can be used to
    evaluate material properties at a point. This function
    allows the user to define the property returned
    as a string. The function can itself be passed to another
    function
    (such as nonlinear_fitting.confidence_prediction_bands()).

    Properties can either be simple attributes (e.g. K_T) or
    a product of attributes, each raised to some power.

    Parameters
    ----------
    m : Material
        The material instance evaluated by the output function.
    attributes : list of strings
        The list of material attributes / properties to
        be evaluated in the product
    powers : list of floats
        The powers to which each attribute should be raised
        during evaluation
    Returns
    -------
    f : function(x)
        Function which returns the value of product(a_i**p_i)
        as a function of condition (x = [P, T, V])
    """
    if type(attributes) is str:
        attributes = [attributes]
    if powers == []:
        powers = [1. for a in attributes]

    def f(x):
        P, T, V = x
        m.set_state(P, T)
        value = 1.
        for a, p in zip(*[attributes, powers]):
            value *= np.power(getattr(m, a), p)
        return value
    return f



from __future__ import print_function
# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

import numpy as np
import warnings
from .material import material_property


class Planet(object):
    """
    A class to build (self-consistent) Planets made out of Layers
    (``burnman.Layer``). By default the planet is set to be self-consistent
    (with zero pressure at the surface and zero gravity at the center),
    but this can be overwritte using the set_pressure_mode().
    Pressure_modes defined in the individual layers will be ignored.
    If temperature modes are already set for each of the layers, when the
    planet is initialized, the planet will be built immediately.
    """

    def __init__(self, name, layers, n_max_iterations=50,
                 max_delta=1.e-5, verbose=False):
        """
        Parameters
        ----------
        name : string
            Name of planet
        layers : list of ``burnman.Layer``
            Layers to build the planet out of
            (layers are sorted within the planet)
        n_max_iterations : int
            Maximum number of iterations to reach self-consistent planet
            (default = 50)
        max_delta : float
            Relative update to the center pressure of the planet between
            iterations to stop iterations (default = 1.e-5)
        """
        # sort layers
        self.layers = sorted(layers, key=lambda x: x.inner_radius)
        # assert layers attach to one another
        if len(self.layers) > 1:
            for i in range(1, len(self.layers)):
                assert(self.layers[i].inner_radius
                       == self.layers[i - 1].outer_radius)

        self.name = name

        self.radii = self.evaluate(['radii'])
        self.n_slices = len(self.radii)
        self.radius_planet = max(self.radii)
        self.volume = 4./3.*np.pi*np.power(self.radius_planet, 3.)

        for layer in self.layers:
            layer.n_start = np.where(self.radii == layer.inner_radius)[0][-1]
            layer.n_end = np.where(self.radii == layer.outer_radius)[0][0] + 1
        self._cached = {}
        self.verbose = verbose
        self.set_pressure_mode(
            n_max_iterations=n_max_iterations, max_delta=max_delta)

    def __iter__(self):
        """
        Planet object will iterate over Layers.
        """
        return list(self.layers).__iter__()

    def __str__(self):
        """
        Prints details of the planet
        """
        writing = '{0} consists of {1} layers:\n'.format(
            self.name, len(self.layers))
        for layer in self:
            writing = writing + layer.__str__()
        return writing

    def reset(self):
        """
        Resets all cached material properties.
        It is typically not required for the user to call this function.
        """
        self._cached = {}

    def get_layer(self, name):
        """
        Returns a layer with a given name

        Parameters
        ----------
        name : string
            Given name of a layer
        """
        for layer in self.layers:
            if layer.name == name:
                return layer
        raise LookupError()

    def get_layer_by_radius(self, radius):
        """
        Returns a layer in which this radius lies

        Parameters
        ----------
        radius : float
            radius at which to evaluate the layer

        """
        for layer in self.layers:
            if layer.outer_radius >= radius:
                return layer
        raise LookupError()

    def evaluate(self, properties, radlist=None):
        """
        Function that is generally used to evaluate properties
        of the different layers and stitch them together.
        If asking for different radii than the internal radlist,
        pressure and temperature values are interpolated and the
        layer material evaluated at those pressures and
        temperatures.

        Parameters
        ----------
        properties : list of strings
            List of properties to evaluate
        radlist : array of floats
            Radii to evaluate properties at. If left empty,
            internal radius lists are used.

        Returns
        -------
        properties_array : numpy array
            1D or 2D array of requested properties
            (1D if only one property was requested)
        """
        if radlist is None:
            values = np.empty([len(properties),
                               np.sum([len(layer.radii)
                                       for layer in self.layers])])
            for i, prop in enumerate(properties):
                if prop == 'depth':
                    values[i] = np.array(
                        [self.radius_planet - r for layer in self.layers
                         for r in layer.radii])
                else:
                    j = 0
                    for layer in self.layers:
                        vals = getattr(layer, prop)
                        values[i, j:j + len(vals)] = vals
                        j += len(vals)
        else:
            values = np.empty([len(properties), len(radlist)])
            l_idx = [i for i, layer in enumerate(self.layers) for r in radlist
                     if r >= layer.inner_radius and r <= layer.outer_radius]

            for j, r in enumerate(radlist):
                values[:, j] = self.layers[l_idx[j]].evaluate(
                    properties, [r], self.radius_planet).T[0]

        if values.shape[0] == 1:
            values = values[0]
        return values

    def set_pressure_mode(self, pressure_mode='self-consistent',
                          pressures=None, pressure_top=0.,
                          gravity_bottom=0., n_max_iterations=50,
                          max_delta=1.e-5):
        """
        Sets the pressure mode of the planet by user-defined values are in a
        self-consistent fashion.
        pressure_mode is 'user-defined' or 'self-consistent'.
        The default for the planet is self-consistent, with zero pressure at
        the surface and zero pressure at the center.

        Parameters
        ----------
        pressure_mode : string
            This can be set to 'user-defined' or 'self-consistent'
        pressures : array of floats
            Pressures (Pa) to set layer to ('user-defined'). This should be the
            same length as defined radius array for the layer
        pressure_top : float
            Pressure (Pa) at the top of the layer.
        gravity_bottom : float
            gravity (m/s^2) at the bottom the layer
        n_max_iterations : int
            Maximum number of iterations to reach self-consistent pressures
            (default = 50)
        """
        self.reset()
        assert(pressure_mode == 'user-defined' or pressure_mode == 'self-consistent')

        self.pressure_mode = pressure_mode
        self.gravity_bottom = gravity_bottom

        if pressure_mode == 'user-defined':
            assert(len(pressures) == len(self.radii))
            self._pressures = pressures
            warnings.warn('User-defined pressures mean that the planet is '
                          'unlikely to be self-consistent')

        if pressure_mode == 'self-consistent':
            self.pressure_top = pressure_top
            self.n_max_iterations = n_max_iterations
            self.max_delta = max_delta

    def make(self):
        """
        This routine needs to be called before evaluating any properties. If pressures and temperatures are self-consistent, they
        are computed across the planet here. Also initializes an array of materials in each Layer to compute properties from.
        """

        self.reset()
        for layer in self.layers:
            assert(layer.temperature_mode is not None)

        if self.pressure_mode == 'user-defined':
            self._temperatures = self._evaluate_temperature(self._pressures)

        if self.pressure_mode == 'self-consistent':

            new_press = self.pressure_top + \
                (-self.radii + max(self.radii)) * \
                1.e3  # initial pressure curve guess
            temperatures = self._evaluate_temperature(new_press)

            # Make it self-consistent!!!
            i = 0
            while i < self.n_max_iterations:
                i += 1
                ref_press = new_press
                new_grav, new_press = self._evaluate_eos(new_press,
                                                         temperatures,
                                                         self.gravity_bottom,
                                                         self.pressure_top)
                temperatures = self._evaluate_temperature(new_press)
                rel_err = abs((max(ref_press) - max(new_press))
                              / max(new_press))
                if self.verbose:
                    print(f'Iteration {i:0d} maximum relative pressure error: '
                          f'{rel_err:.1e}')

                if rel_err < self.max_delta:
                    break

            self.pressures = new_press
            self.temperatures = temperatures
            self._gravity = new_grav

        for layer in self.layers:
            layer.sublayers = []
            layer.pressures = self.pressures[layer.n_start: layer.n_end]
            layer.temperatures = self.temperatures[layer.n_start: layer.n_end]
            layer.gravity_bottom = self._gravity[layer.n_start - 1]
            layer.pressure_mode = 'set-in-planet'
            for i in range(len(layer.radii)):
                layer.sublayers.append(layer.material.copy())
                layer.sublayers[i].set_state(
                    layer.pressures[i], layer.temperatures[i])

    def _evaluate_eos(self, pressures, temperatures,
                      gravity_bottom, pressure_top):
        """
        Used to update the pressure profile in set_state()
        """
        density = self._evaluate_density(pressures, temperatures)
        grav = self._compute_gravity(density, gravity_bottom)
        press = self._compute_pressure(density, grav, pressure_top)
        return grav, press

    def _evaluate_density(self, pressures, temperatures):
        """
        Used to update the density profile in _evaluate_eos()
        """
        density = []
        for layer in self.layers:
            density.append(layer.material.evaluate(['density'],
                                                   pressures[layer.n_start:layer.n_end],
                                                   temperatures[layer.n_start:layer.n_end]))
        return np.squeeze(np.hstack(density))

    def _evaluate_temperature(self, pressures):
        """
        Returns the temperatures of different layers for given pressures.
        Used by set_state()
        """
        temps = []
        temperature_top = None
        for layer in self.layers[::-1]:
            if temperature_top is None or layer.temperature_top is not None:
                temperature_top = layer.temperature_top
            temps.extend(layer._evaluate_temperature(
                (pressures[layer.n_start:layer.n_end]), temperature_top)[::-1])
            temperature_top = temps[-1]
        return np.hstack(np.squeeze(temps))[::-1]

    def _compute_gravity(self, density, gravity_bottom):
        """
        Calculate the gravity of the planet, based on a density profile.
        This integrates Poisson's equation in radius, under the assumption
        that the planet is laterally homogeneous.
        Used to update the gravity profile in _evaluate_eos()
        """

        start_gravity = gravity_bottom
        grav = []
        for layer in self.layers:
            grav.extend(layer._compute_gravity(
                density[layer.n_start: layer.n_end], start_gravity))
            start_gravity = grav[-1]

        return np.array(grav)

    def _compute_pressure(self, density, gravity, pressure_top):
        """
        Calculate the pressure profile based on density and gravity.
        This integrates the equation for hydrostatic equilibrium P = rho g z.
        Used to update the pressure profile in _evaluate_eos()
        """
        start_pressure = pressure_top
        press = []
        for layer in self.layers[::-1]:
            press.extend(layer._compute_pressure(density[layer.n_start:layer.n_end],
                                                 gravity[layer.n_start: layer.n_end],
                                                 start_pressure)[::-1])
            start_pressure = press[-1]
        return np.array(press)[::-1]

    @property
    def mass(self):
        """
        calculates the mass of the entire planet [kg]
        """
        return np.sum([layer.mass for layer in self.layers])

    @property
    def average_density(self):
        """
        calculates the average density of the entire planet [kg/m^3]
        """
        return self.mass/self.volume

    @property
    def moment_of_inertia(self):
        """
        #Returns the moment of inertia of the planet [kg m^2]
        """
        return np.sum([layer.moment_of_inertia for layer in self.layers])

    @property
    def moment_of_inertia_factor(self):
        """
        #Returns the moment of inertia of the planet [kg m^2]
        """
        moment_factor = self.moment_of_inertia / self.mass / \
            self.radius_planet / self.radius_planet
        return moment_factor

    @property
    def depth(self):
        """
        Returns depth of the layer [m]
        """
        return self.evaluate(['depth'])

    @property
    def gravity(self):
        """
        Returns gravity of the layer [m s^(-2)]
        """
        return self.evaluate(['gravity'])

    @property
    def bullen(self):
        """
        Returns the Bullen parameter
        """
        return self.evaluate(['bullen'])

    @property
    def brunt_vasala(self):
        return self.evaluate(['brunt_vasala'])

    @property
    def pressure(self):
        """
        Returns current pressure that was set with
        :func:`~burnman.Material.set_state`.

        Aliased with :func:`~burnman.Material.P`.

        Returns
        -------
        pressure : array of floats
            Pressure in [Pa].
        """
        return self.pressures

    @property
    def temperature(self):
        """
        Returns current temperature that was set with
        :func:`~burnman.Material.set_state`.

        Aliased with :func:`~burnman.Material.T`.

        Returns
        -------
        temperature : array of floats
            Temperature in [K].
        """
        return self.temperatures

    @material_property
    def molar_internal_energy(self):
        """
        Returns the molar internal energy of the planet.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.energy`.

        Returns
        -------
        molar_internal_energy : array of floats
            The internal energy in [J/mol].
        """
        return self.evaluate(['molar_internal_energy'])

    @material_property
    def molar_gibbs(self):
        """
        Returns the molar Gibbs free energy of the planet.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.gibbs`.

        Returns
        -------
        molar_gibbs : array of floats
            Gibbs free energy in [J/mol].
        """
        return self.evaluate(['molar_gibbs'])

    @material_property
    def molar_helmholtz(self):
        """
        Returns the molar Helmholtz free energy of the planet.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.helmholtz`.

        Returns
        -------
        molar_helmholtz : array of floats
            Helmholtz free energy in [J/mol].
        """
        return self.evaluate(['molar_helmholtz'])

    @material_property
    def molar_mass(self):
        """
        Returns molar mass of the planet.

        Needs to be implemented in derived classes.

        Returns
        -------
        molar_mass : array of floats
            Molar mass in [kg/mol].
        """
        return self.evaluate(['molar_mass'])

    @material_property
    def molar_volume(self):
        """
        Returns molar volume of the planet.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.V`.

        Returns
        -------
        molar_volume : array of floats
            Molar volume in [m^3/mol].
        """
        return self.evaluate(['molar_volume'])

    @material_property
    def density(self):
        """
        Returns the density of this planet.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.rho`.

        Returns
        -------
        density : array of floats
            The density of this material in [kg/m^3].
        """
        return self.evaluate(['density'])

    @material_property
    def molar_entropy(self):
        """
        Returns molar entropy of the planet.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.S`.

        Returns
        -------
        molar_entropy : array of floats
            Entropy in [J/mol].
        """
        return self.evaluate(['molar_entropy'])

    @material_property
    def molar_enthalpy(self):
        """
        Returns molar enthalpy of the planet.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.H`.

        Returns
        -------
        molar_enthalpy : array of floats
            Enthalpy in [J/mol].
        """
        return self.evaluate(['molar_enthalpy'])

    @material_property
    def isothermal_bulk_modulus(self):
        """
        Returns isothermal bulk modulus of the planet.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.K_T`.

        Returns
        -------
        isothermal_bulk_modulus : array of floats
            Bulk modulus in [Pa].
        """
        return self.evaluate(['isothermal_bulk_modulus'])

    @material_property
    def adiabatic_bulk_modulus(self):
        """
        Returns the adiabatic bulk modulus of the planet.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.K_S`.

        Returns
        -------
        adiabatic_bulk_modulus : array of floats
            Adiabatic bulk modulus in [Pa].
        """
        return self.evaluate(['adiabatic_bulk_modulus'])

    @material_property
    def isothermal_compressibility(self):
        """
        Returns isothermal compressibility of the planet
        (or inverse isothermal bulk modulus).

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.beta_T`.

        Returns
        -------
        (K_T)^-1 : array of floats
            Compressibility in [1/Pa].
        """
        return self.evaluate(['istothermal_compressibility'])

    @material_property
    def adiabatic_compressibility(self):
        """
        Returns adiabatic compressibility of the planet
        (or inverse adiabatic bulk modulus).

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.beta_S`.

        Returns
        -------
        adiabatic_compressibility : array of floats
            adiabatic compressibility in [1/Pa].
        """
        return self.evaluate(['adiabatic_compressibility'])

    @material_property
    def shear_modulus(self):
        """
        Returns shear modulus of the planet.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.beta_G`.

        Returns
        -------
        shear_modulus : array of floats
            Shear modulus in [Pa].
        """
        return self.evaluate(['shear_modulus'])

    @material_property
    def p_wave_velocity(self):
        """
        Returns P wave speed of the planet.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.v_p`.

        Returns
        -------
        p_wave_velocity : array of floats
            P wave speed in [m/s].
        """
        return self.evaluate(['p_wave_velocity'])

    @material_property
    def bulk_sound_velocity(self):
        """
        Returns bulk sound speed of the planet.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.v_phi`.

        Returns
        -------
        bulk sound velocity: array of floats
            Sound velocity in [m/s].
        """
        return self.evaluate(['bulk_sound_velocity'])

    @material_property
    def shear_wave_velocity(self):
        """
        Returns shear wave speed of the planet.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.v_s`.

        Returns
        -------
        shear_wave_velocity : array of floats
            Wave speed in [m/s].
        """
        return self.evaluate(['shear_wave_velocity'])

    @material_property
    def grueneisen_parameter(self):
        """
        Returns the grueneisen parameter of the planet.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.gr`.

        Returns
        -------
        gr : array of floats
            Grueneisen parameters [unitless].
        """
        return self.evaluate(['grueneisen_parameter'])

    @material_property
    def thermal_expansivity(self):
        """
        Returns thermal expansion coefficient of the planet.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.alpha`.

        Returns
        -------
        alpha : array of floats
            Thermal expansivity in [1/K].
        """
        return self.evaluate(['thermal_expansivity'])

    @material_property
    def molar_heat_capacity_v(self):
        """
        Returns molar heat capacity at constant volume of the planet.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.C_v`.

        Returns
        -------
        molar_heat_capacity_v : array of floats
            Heat capacity in [J/K/mol].
        """
        return self.evaluate(['molar_heat_capacity_v'])

    @material_property
    def molar_heat_capacity_p(self):
        """
        Returns molar heat capacity at constant pressure of the planet.

        Needs to be implemented in derived classes.
        Aliased with :func:`~burnman.Material.C_p`.

        Returns
        -------
        molar_heat_capacity_p : array of floats
            Heat capacity in [J/K/mol].
        """
        return self.evaluate(['molar_heat_capacity_p'])

#
# Aliased properties
    @property
    def P(self):
        """Alias for :func:`~burnman.Material.pressure`"""
        return self.pressure

    @property
    def T(self):
        """Alias for :func:`~burnman.Material.temperature`"""
        return self.temperature

    @property
    def energy(self):
        """Alias for :func:`~burnman.Material.molar_internal_energy`"""
        return self.molar_internal_energy

    @property
    def helmholtz(self):
        """Alias for :func:`~burnman.Material.molar_helmholtz`"""
        return self.molar_helmholtz

    @property
    def gibbs(self):
        """Alias for :func:`~burnman.Material.molar_gibbs`"""
        return self.molar_gibbs

    @property
    def V(self):
        """Alias for :func:`~burnman.Material.molar_volume`"""
        return self.molar_volume

    @property
    def rho(self):
        """Alias for :func:`~burnman.Material.density`"""
        return self.density

    @property
    def S(self):
        """Alias for :func:`~burnman.Material.molar_entropy`"""
        return self.molar_entropy

    @property
    def H(self):
        """Alias for :func:`~burnman.Material.molar_enthalpy`"""
        return self.molar_enthalpy

    @property
    def K_T(self):
        """Alias for :func:`~burnman.Material.isothermal_bulk_modulus`"""
        return self.isothermal_bulk_modulus

    @property
    def K_S(self):
        """Alias for :func:`~burnman.Material.adiabatic_bulk_modulus`"""
        return self.adiabatic_bulk_modulus

    @property
    def beta_T(self):
        """Alias for :func:`~burnman.Material.isothermal_compressibility`"""
        return self.isothermal_compressibility

    @property
    def beta_S(self):
        """Alias for :func:`~burnman.Material.adiabatic_compressibility`"""
        return self.adiabatic_compressibility

    @property
    def G(self):
        """Alias for :func:`~burnman.Material.shear_modulus`"""
        return self.shear_modulus

    @property
    def v_p(self):
        """Alias for :func:`~burnman.Material.p_wave_velocity`"""
        return self.p_wave_velocity

    @property
    def v_phi(self):
        """Alias for :func:`~burnman.Material.bulk_sound_velocity`"""
        return self.bulk_sound_velocity

    @property
    def v_s(self):
        """Alias for :func:`~burnman.Material.shear_wave_velocity`"""
        return self.shear_wave_velocity

    @property
    def gr(self):
        """Alias for :func:`~burnman.Material.grueneisen_parameter`"""
        return self.grueneisen_parameter

    @property
    def alpha(self):
        """Alias for :func:`~burnman.Material.thermal_expansivity`"""
        return self.thermal_expansivity

    @property
    def C_v(self):
        """Alias for :func:`~burnman.Material.molar_heat_capacity_v`"""
        return self.molar_heat_capacity_v

    @property
    def C_p(self):
        """Alias for :func:`~burnman.Material.molar_heat_capacity_p`"""
        return self.molar_heat_capacity_p
